#!/usr/bin/env python3

# This file generates the member definitions and declarations for all vips
# operators.

# this needs pyvips
#   pip install --user pyvips

# rebuild with:
#   meson compile -Cbuild vips-operators-header
#   meson compile -Cbuild vips-operators-source

# Sample member declaration:
# VImage invert(VOption *options = nullptr) const;

# Sample member definition:
# VImage
# VImage::invert(VOption *options) const
# {
#   VImage out;
#
# 	call("invert", (options ? options : VImage::option())
# 				->set("in", *this)
# 				->set("out", &out));
#
# 	return out;
# }

import argparse

from pyvips import Introspect, Operation, GValue, Error, \
    ffi, gobject_lib, type_map, type_from_name, nickname_find, type_name

# turn a GType into a C++ type
gtype_to_cpp = {
    GValue.gbool_type: 'bool',
    GValue.gint_type: 'int',
    GValue.gdouble_type: 'double',
    GValue.gstr_type: 'const char *',
    GValue.refstr_type: 'char *',
    GValue.image_type: 'VImage',
    GValue.source_type: 'VSource',
    GValue.target_type: 'VTarget',
    GValue.guint64_type: 'guint64',
    type_from_name('VipsInterpolate'): 'VInterpolate',
    GValue.array_int_type: 'std::vector<int>',
    GValue.array_double_type: 'std::vector<double>',
    GValue.array_image_type: 'std::vector<VImage>',
    GValue.blob_type: 'VipsBlob *'
}

cplusplus_suffixes = ('*', '&')
cplusplus_keywords = ('case', 'switch')

# values for VipsArgumentFlags
_REQUIRED = 1
_INPUT = 16
_OUTPUT = 32
_DEPRECATED = 64
_MODIFY = 128

# for VipsOperationFlags
_OPERATION_NOCACHE = 4
_OPERATION_DEPRECATED = 8


def get_cpp_type(gtype):
    """Map a gtype to the C++ type name we use to represent it.
    """
    if gtype in gtype_to_cpp:
        return gtype_to_cpp[gtype]

    fundamental = gobject_lib.g_type_fundamental(gtype)

    # enum/flag params use the C name as their name
    if fundamental == GValue.genum_type or fundamental == GValue.gflags_type:
        return type_name(gtype)

    if fundamental in gtype_to_cpp:
        return gtype_to_cpp[fundamental]

    return '<unknown type>'


# swap any '-' for '_'
def cppize(name):
    return name.replace('-', '_')


def generate_operation(operation_name, declaration_only=False, indent=''):
    intro = Introspect.get(operation_name)

    required_output = [name for name in intro.required_output if
                       name != intro.member_x]

    # We are only interested in non-deprecated arguments
    optional_input = [name for name in intro.optional_input if
                      intro.details[name]['flags'] & _DEPRECATED == 0]

    # Drop "revalidate" flag from operations marked "nocache"
    optional_input = [name for name in optional_input if
                      name != 'revalidate' or (intro.flags & _OPERATION_NOCACHE) == 0]

    has_output = len(required_output) >= 1

    # Add a C++ style comment block with some additional markings (@param,
    # @return)
    if declaration_only:
        description = intro.description.capitalize()
        # hide library-specific implementation details
        description = (description
                       .replace('imagemagick7', 'imagemagick')
                       .replace(' (pdfium)', '')
                       .replace(' (poppler)', ''))
        result = f'\n{indent}/**'
        result += f'\n{indent} * {description}.'

        if len(optional_input) > 0:
            result += f'\n{indent} *'
            result += f'\n{indent} * **Optional parameters**'
            for name in optional_input:
                details = intro.details[name]
                result += f'\n{indent} *   - **{cppize(name)}** -- '
                result += f'{details["blurb"]}, '
                result += f'{get_cpp_type(details["type"])}.'
            result += f'\n{indent} *'

        for name in intro.method_args:
            details = intro.details[name]
            result += f'\n{indent} * @param {cppize(name)} {details["blurb"]}.'

        if has_output:
            # skip the first element
            for name in required_output[1:]:
                details = intro.details[name]
                result += f'\n{indent} * @param {cppize(name)} {details["blurb"]}.'

        result += f'\n{indent} * @param options Set of options.'

        if has_output:
            details = intro.details[required_output[0]]
            result += f'\n{indent} * @return {details["blurb"]}.'

        result += f'\n{indent} */\n'
        if intro.flags & _OPERATION_DEPRECATED:
            result += f'{indent}G_DEPRECATED\n'
    else:
        result = '\n'

    if intro.member_x is None and declaration_only:
        result += f'{indent}static '
    else:
        result += indent
    if has_output:
        # the first output arg will be used as the result
        cpp_type = get_cpp_type(intro.details[required_output[0]]['type'])
        result += f'{cpp_type}'
    else:
        cpp_type = 'void'
        result += cpp_type

    if declaration_only:
        result += '' if cpp_type.endswith(cplusplus_suffixes) else ' '
    else:
        result += '\nVImage::'

    cplusplus_operation = operation_name
    if operation_name in cplusplus_keywords:
        cplusplus_operation += '_image'

    result += f'{cplusplus_operation}('
    for name in intro.method_args:
        details = intro.details[name]
        gtype = details['type']
        cpp_type = get_cpp_type(gtype)
        spacing = '' if cpp_type.endswith(cplusplus_suffixes) else ' '
        result += f'{cpp_type}{spacing}{cppize(name)}, '

    # output params are passed by reference
    if has_output:
        # skip the first element
        for name in required_output[1:]:
            details = intro.details[name]
            gtype = details['type']
            cpp_type = get_cpp_type(gtype)
            spacing = '' if cpp_type.endswith(cplusplus_suffixes) else ' '
            result += f'{cpp_type}{spacing}*{cppize(name)}, '

    result += f'VOption *options{" = nullptr" if declaration_only else ""})'

    # if no 'this' available, it's a class method and they are all const
    if intro.member_x is not None:
        result += ' const'

    if declaration_only:
        result += ';'

        return result

    result += '\n{\n'

    if has_output:
        # the first output arg will be used as the result
        name = required_output[0]
        cpp_type = get_cpp_type(intro.details[name]['type'])
        spacing = '' if cpp_type.endswith(cplusplus_suffixes) else ' '
        result += f'\t{cpp_type}{spacing}{cppize(name)};\n\n'

    result += f'\tcall("{operation_name}", (options ? options : VImage::option())'
    if intro.member_x is not None:
        result += f'\n\t\t\t->set("{intro.member_x}", *this)'

    all_required = intro.method_args

    if has_output:
        # first element needs to be passed by reference
        arg = cppize(required_output[0])
        result += f'\n\t\t\t->set("{required_output[0]}", &{arg})'

        # append the remaining list
        all_required += required_output[1:]

    for name in all_required:
        arg = cppize(name)
        result += f'\n\t\t\t->set("{name}", {arg})'

    result += ');\n'

    if has_output:
        result += f'\n'
        result += f'\treturn {required_output[0]};\n'

    result += '}'

    return result


def generate_operators(declarations_only=False):
    all_nicknames = []

    # hidden CLI savers needs to be excluded from the API
    hidden_savers = [
        'avifsave_target',
        'magicksave_bmp',
        'magicksave_bmp_buffer',
        'pbmsave_target',
        'pfmsave_target',
        'pgmsave_target',
        'pnmsave_target',
    ]

    def add_nickname(gtype, a, b):
        nickname = nickname_find(gtype)
        try:
            # can fail for abstract types
            _ = Introspect.get(nickname)

            all_nicknames.append(nickname)
        except Error:
            pass

        type_map(gtype, add_nickname)

        return ffi.NULL

    type_map(type_from_name('VipsOperation'), add_nickname)

    # add 'missing' synonyms by hand
    all_nicknames.append('crop')

    # make list unique and sort
    all_nicknames = list(set(all_nicknames) - set(hidden_savers))
    all_nicknames.sort()

    indent = '\t' if declarations_only else ''

    print(f'''{indent}// {'headers' if declarations_only else 'bodies'} for vips operations
{indent}// this file is generated automatically, do not edit!
{indent}// clang-format off''')

    for nickname in all_nicknames:
        print(generate_operation(nickname, declarations_only, indent))


parser = argparse.ArgumentParser(description='C++ binding generator')
parser.add_argument('--gen', '-g',
                    default='cpp',
                    choices=['h', 'cpp'],
                    help='File to generate: h (headers) or cpp ' + \
                         '(implementations) (default: %(default)s)')

if __name__ == '__main__':
    args = parser.parse_args()

    generate_operators(args.gen == 'h')



# vim: set fileencoding=utf-8 :
import pytest

import pyvips
from helpers import *


class TestCreate:
    def test_black(self):
        im = pyvips.Image.black(100, 100)

        assert im.width == 100
        assert im.height == 100
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.bands == 1
        for i in range(0, 100):
            pixel = im(i, i)
            assert len(pixel) == 1
            assert pixel[0] == 0

        im = pyvips.Image.black(100, 100, bands=3)

        assert im.width == 100
        assert im.height == 100
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.bands == 3
        for i in range(0, 100):
            pixel = im(i, i)
            assert len(pixel) == 3
            assert_almost_equal_objects(pixel, [0, 0, 0])

    def test_buildlut(self):
        M = pyvips.Image.new_from_array([[0, 0],
                                         [255, 100]])
        lut = M.buildlut()
        assert lut.width == 256
        assert lut.height == 1
        assert lut.bands == 1
        p = lut(0, 0)
        assert p[0] == 0.0
        p = lut(255, 0)
        assert p[0] == 100.0
        p = lut(10, 0)
        assert p[0] == 100 * 10.0 / 255.0

        M = pyvips.Image.new_from_array([[0, 0, 100],
                                         [255, 100, 0],
                                         [128, 10, 90]])
        lut = M.buildlut()
        assert lut.width == 256
        assert lut.height == 1
        assert lut.bands == 2
        p = lut(0, 0)
        assert_almost_equal_objects(p, [0.0, 100.0])
        p = lut(64, 0)
        assert_almost_equal_objects(p, [5.0, 95.0])

    def test_eye(self):
        im = pyvips.Image.eye(100, 90)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.max() == 1.0
        assert im.min() == -1.0

        im = pyvips.Image.eye(100, 90, uchar=True)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.max() == 255.0
        assert im.min() == 0.0

    @skip_if_no("fwfft")
    def test_fwfft_small_image(self):
        im = pyvips.Image.black(2, 1)
        im.fwfft()

    @skip_if_no("fwfft")
    def test_fractsurf(self):
        im = pyvips.Image.fractsurf(100, 90, 2.5)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

    def test_gaussmat(self):
        im = pyvips.Image.gaussmat(1, 0.1)
        assert im.width == 5
        assert im.height == 5
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.DOUBLE
        assert im.max() == 20
        total = im.avg() * im.width * im.height
        scale = im.get("scale")
        assert total == scale
        p = im(im.width / 2, im.height / 2)
        assert p[0] == 20.0

        im = pyvips.Image.gaussmat(1, 0.1,
                                   separable=True, precision="float")
        assert im.width == 5
        assert im.height == 1
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.DOUBLE
        assert im.max() == 1.0
        total = im.avg() * im.width * im.height
        scale = im.get("scale")
        assert total == scale
        p = im(im.width / 2, im.height / 2)
        assert p[0] == 1.0

    def test_gaussnoise(self):
        im = pyvips.Image.gaussnoise(100, 90)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

        im = pyvips.Image.gaussnoise(100, 90, sigma=10, mean=100)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

        sigma = im.deviate()
        mean = im.avg()

        assert sigma == pytest.approx(10, abs=0.4)
        assert mean == pytest.approx(100, abs=0.4)

    def test_grey(self):
        im = pyvips.Image.grey(100, 90)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

        p = im(0, 0)
        assert p[0] == 0.0
        p = im(99, 0)
        assert p[0] == 1.0
        p = im(0, 89)
        assert p[0] == 0.0
        p = im(99, 89)
        assert p[0] == 1.0

        im = pyvips.Image.grey(100, 90, uchar=True)
        assert im.width == 100
        assert im.height == 90
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR

        p = im(0, 0)
        assert p[0] == 0
        p = im(99, 0)
        assert p[0] == 255
        p = im(0, 89)
        assert p[0] == 0
        p = im(99, 89)
        assert p[0] == 255

    def test_identity(self):
        im = pyvips.Image.identity()
        assert im.width == 256
        assert im.height == 1
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR

        p = im(0, 0)
        assert p[0] == 0.0
        p = im(255, 0)
        assert p[0] == 255.0
        p = im(128, 0)
        assert p[0] == 128.0

        im = pyvips.Image.identity(ushort=True)
        assert im.width == 65536
        assert im.height == 1
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.USHORT

        p = im(0, 0)
        assert p[0] == 0
        p = im(99, 0)
        assert p[0] == 99
        p = im(65535, 0)
        assert p[0] == 65535

    def test_invertlut(self):
        lut = pyvips.Image.new_from_array([[0.1, 0.2, 0.3, 0.1],
                                           [0.2, 0.4, 0.4, 0.2],
                                           [0.7, 0.5, 0.6, 0.3]])
        im = lut.invertlut()
        assert im.width == 256
        assert im.height == 1
        assert im.bands == 3
        assert im.format == pyvips.BandFormat.DOUBLE

        p = im(0, 0)
        assert_almost_equal_objects(p, [0, 0, 0])
        p = im(255, 0)
        assert_almost_equal_objects(p, [1, 1, 1])
        p = im(0.2 * 255, 0)
        assert p[0] == pytest.approx(0.1, abs=0.1)
        p = im(0.3 * 255, 0)
        assert p[1] == pytest.approx(0.1, abs=0.1)
        p = im(0.1 * 255, 0)
        assert p[2] == pytest.approx(0.1, abs=0.1)

    def test_matrixinvert(self):
        # 4x4 matrix to check if PLU decomposition works
        mat = pyvips.Image.new_from_array([[4, 0, 0, 0],
                                           [0, 0, 2, 0],
                                           [0, 1, 2, 0],
                                           [1, 0, 0, 1]])
        im = mat.matrixinvert()
        assert im.width == 4
        assert im.height == 4
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.DOUBLE

        p = im(0, 0)
        assert p[0] == 0.25
        p = im(3, 3)
        assert p[0] == 1.0

    def test_logmat(self):
        im = pyvips.Image.logmat(1, 0.1)
        assert im.width == 7
        assert im.height == 7
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.DOUBLE
        assert im.max() == 20
        total = im.avg() * im.width * im.height
        scale = im.get("scale")
        assert total == scale
        p = im(im.width / 2, im.height / 2)
        assert p[0] == 20.0

        im = pyvips.Image.logmat(1, 0.1,
                                 separable=True, precision="float")
        assert im.width == 7
        assert im.height == 1
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.DOUBLE
        assert im.max() == 1.0
        total = im.avg() * im.width * im.height
        scale = im.get("scale")
        assert total == scale
        p = im(im.width / 2, im.height / 2)
        assert p[0] == 1.0

    def test_mask_butterworth_band(self):
        im = pyvips.Image.mask_butterworth_band(128, 128, 2,
                                                0.5, 0.5, 0.7,
                                                0.1)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.max() == pytest.approx(1, abs=0.01)
        p = im(32, 32)
        assert p[0] == 1.0

        im = pyvips.Image.mask_butterworth_band(128, 128, 2,
                                                0.5, 0.5, 0.7,
                                                0.1, uchar=True, optical=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.max() == 255
        p = im(32, 32)
        assert p[0] == 255.0
        p = im(64, 64)
        assert p[0] == 255.0

        im = pyvips.Image.mask_butterworth_band(128, 128, 2,
                                                0.5, 0.5, 0.7,
                                                0.1, uchar=True, optical=True,
                                                nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.max() == 255
        p = im(32, 32)
        assert p[0] == 255.0
        p = im(64, 64)
        assert p[0] != 255

    def test_mask_butterworth(self):
        im = pyvips.Image.mask_butterworth(128, 128, 2, 0.7, 0.1,
                                           nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.min() == pytest.approx(0, abs=0.01)
        p = im(0, 0)
        assert p[0] == 0.0
        v, x, y = im.maxpos()
        assert x == 64
        assert y == 64

        im = pyvips.Image.mask_butterworth(128, 128, 2, 0.7, 0.1,
                                           optical=True, uchar=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.min() == pytest.approx(0, abs=0.01)
        p = im(64, 64)
        assert p[0] == 255

    def test_mask_butterworth_ring(self):
        im = pyvips.Image.mask_butterworth_ring(128, 128, 2, 0.7, 0.1, 0.5,
                                                nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        p = im(45, 0)
        assert p[0] == pytest.approx(1.0, abs=0.0001)
        v, x, y = im.minpos()
        assert x == 64
        assert y == 64

    def test_mask_fractal(self):
        im = pyvips.Image.mask_fractal(128, 128, 2.3)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

    def test_mask_gaussian_band(self):
        im = pyvips.Image.mask_gaussian_band(128, 128, 0.5, 0.5, 0.7, 0.1)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.max() == pytest.approx(1, abs=0.01)
        p = im(32, 32)
        assert p[0] == 1.0

    def test_mask_gaussian(self):
        im = pyvips.Image.mask_gaussian(128, 128, 0.7, 0.1,
                                        nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.min() == pytest.approx(0, abs=0.01)
        p = im(0, 0)
        assert p[0] == 0.0

    def test_mask_gaussian_ring(self):
        im = pyvips.Image.mask_gaussian_ring(128, 128, 0.7, 0.1, 0.5,
                                             nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        p = im(45, 0)
        assert p[0] == pytest.approx(1.0, abs=0.001)

    def test_mask_ideal_band(self):
        im = pyvips.Image.mask_ideal_band(128, 128, 0.5, 0.5, 0.7)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.max() == pytest.approx(1, abs=0.01)
        p = im(32, 32)
        assert p[0] == 1.0

    def test_mask_ideal(self):
        im = pyvips.Image.mask_ideal(128, 128, 0.7,
                                     nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.min() == pytest.approx(0, abs=0.01)
        p = im(0, 0)
        assert p[0] == 0.0

    def test_mask_gaussian_ring_2(self):
        im = pyvips.Image.mask_ideal_ring(128, 128, 0.7, 0.5,
                                          nodc=True)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        p = im(45, 0)
        assert p[0] == pytest.approx(1, abs=0.001)

    def test_sines(self):
        im = pyvips.Image.sines(128, 128)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

    @skip_if_no("text")
    def test_text(self):
        im = pyvips.Image.text("Hello, world!", dpi=300)
        assert im.width > 10
        assert im.height > 10
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.UCHAR
        assert im.max() > 240
        assert im.min() == 0

        # test autofit
        im = pyvips.Image.text("Hello, world!", width=500, height=500)
        # quite a large threshold, since we need to work with a huge range of
        # text rendering systems
        assert abs(im.width - 500) < 50

        # test wrap
        im1 = pyvips.Image.text("helloworld", width=100, dpi=500)
        im2 = pyvips.Image.text("helloworld", width=100, dpi=500, wrap="char")
        assert im1.width > im2.width

    def test_tonelut(self):
        im = pyvips.Image.tonelut()
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.USHORT
        assert im.width == 32768
        assert im.height == 1
        assert im.hist_ismonotonic()

    def test_xyz(self):
        im = pyvips.Image.xyz(128, 128)
        assert im.bands == 2
        assert im.format == pyvips.BandFormat.UINT
        assert im.width == 128
        assert im.height == 128
        p = im(45, 35)
        assert_almost_equal_objects(p, [45, 35])

    def test_sdf(self):
        im = pyvips.Image.sdf(128, 128, "circle", a=[64, 64], r=32)
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.width == 128
        assert im.height == 128
        p = im(45, 35)
        assert_almost_equal_objects(p, [2.670], threshold=0.01)

        im = pyvips.Image.sdf(128, 128, "box", a=[10, 10], b=[50, 40])
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.width == 128
        assert im.height == 128
        p = im(45, 35)
        assert_almost_equal_objects(p, [-5.0])

        im = pyvips.Image.sdf(128, 128, "rounded-box",
                              a=[10, 10],
                              b=[50, 40],
                              corners=[50, 0, 0, 0])
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.width == 128
        assert im.height == 128
        p = im(45, 35)
        assert_almost_equal_objects(p, [13.640], threshold=0.01)

        im = pyvips.Image.sdf(128, 128, "line", a=[10, 10], b=[50, 40])
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT
        assert im.width == 128
        assert im.height == 128
        p = im(45, 35)
        assert_almost_equal_objects(p, [1.0])

    def test_zone(self):
        im = pyvips.Image.zone(128, 128)
        assert im.width == 128
        assert im.height == 128
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

    def test_worley(self):
        im = pyvips.Image.worley(512, 512)
        assert im.width == 512
        assert im.height == 512
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT

    def test_perlin(self):
        im = pyvips.Image.perlin(512, 512)
        assert im.width == 512
        assert im.height == 512
        assert im.bands == 1
        assert im.format == pyvips.BandFormat.FLOAT


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python3

import argparse
import xml.etree.ElementTree as ET
from pathlib import Path


def register_all_namespaces(filename):
    namespaces = dict([node for _, node in ET.iterparse(filename, events=['start-ns'])])
    for ns in namespaces:
        ET.register_namespace(ns, namespaces[ns])


def fixup_gir_for_doc(args):
    tree = ET.parse(args.in_path)
    root = tree.getroot()

    register_all_namespaces(args.in_path)
    namespace = {
        'goi': 'http://www.gtk.org/introspection/core/1.0'
    }

    namespace_node = root.find('goi:namespace', namespace)
    image_node = namespace_node.find('goi:class[@name="Image"]', namespace)
    blob_node = namespace_node.find('goi:record[@name="Blob"]', namespace)

    # A list of functions that needs to be converted to an Image constructor.
    # Note that all functions containing "load" are already converted.
    image_ctors = [
        'black',
        'eye',
        'fractsurf',
        'gaussmat',
        'gaussnoise',
        'grey',
        'identity',
        'logmat',
        'mask_butterworth',
        'mask_butterworth_band',
        'mask_butterworth_ring',
        'mask_fractal',
        'mask_gaussian',
        'mask_gaussian_band',
        'mask_gaussian_ring',
        'mask_ideal',
        'mask_ideal_band',
        'mask_ideal_ring',
        'perlin',
        'sdf',
        'sines',
        'system',
        'text',
        'thumbnail',
        'thumbnail_buffer',
        'thumbnail_source',
        'tonelut',
        'worley',
        'xyz',
        'zone',
    ]

    # Functions that take multiple images as input argument ... make them
    # part of the Image class.
    # Keep-in-sync with gen-function-list.py
    image_funcs = [
        'arrayjoin',
        'bandjoin',
        'bandrank',
        'composite',
        'sum',
        'switch',
    ]

    for node in namespace_node.findall('goi:function', namespace):
        name = node.get('name')
        if name == 'profile_load':
            namespace_node.remove(node)

            node.tag = 'constructor'
            blob_node.append(node)
        elif 'load' in name or name in image_ctors:
            namespace_node.remove(node)

            node.tag = 'constructor'
            image_node.append(node)
        elif name in image_funcs:
            namespace_node.remove(node)
            image_node.append(node)

    tree.write(args.out_path, encoding='utf-8', xml_declaration=True)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('in_path', help='input file', type=Path)
    parser.add_argument('out_path', help='output file', type=Path)

    args = parser.parse_args()
    fixup_gir_for_doc(args)


if __name__ == '__main__':
    main()



# vim: set fileencoding=utf-8 :
import filecmp
from functools import reduce

import os
import pytest
import tempfile
import shutil

import pyvips
from helpers import *


class TestConversion:
    tempdir = None

    # run a function on an image,
    # 50,50 and 10,10 should have different values on the test image
    # don't loop over band elements
    def run_image_pixels(self, message, im, fn):
        run_cmp(message, im, 50, 50, fn)
        run_cmp(message, im, 10, 10, fn)

    # run a function on a pair of images
    # 50,50 and 10,10 should have different values on the test image
    # don't loop over band elements
    def run_image_pixels2(self, message, left, right, fn):
        run_cmp2(message, left, right, 50, 50, fn)
        run_cmp2(message, left, right, 10, 10, fn)

    def run_unary(self, images, fn, fmt=all_formats):
        [self.run_image_pixels(fn.__name__ + (' %s' % y), x.cast(y), fn)
         for x in images for y in fmt]

    def run_binary(self, images, fn, fmt=all_formats):
        [self.run_image_pixels2(fn.__name__ + (' %s %s' % (y, z)),
                                x.cast(y), x.cast(z), fn)
         for x in images for y in fmt for z in fmt]

    @classmethod
    def setup_class(cls):
        cls.tempdir = tempfile.mkdtemp()
        im = pyvips.Image.mask_ideal(100, 100, 0.5,
                                     reject=True, optical=True)
        cls.colour = (im * [1, 2, 3] + [2, 3, 4]).copy(interpretation="srgb")
        cls.mono = cls.colour[1].copy(interpretation="b-w")
        cls.all_images = [cls.mono, cls.colour]
        cls.image = pyvips.Image.jpegload(JPEG_FILE)

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.tempdir, ignore_errors=True)
        cls.colour = None
        cls.mono = None
        cls.image = None
        cls.all_images = None

    def test_cast(self):
        # casting negative pixels to an unsigned format should clip to zero
        for signed in signed_formats:
            im = (pyvips.Image.black(1, 1) - 10).cast(signed)
            for unsigned in unsigned_formats:
                im2 = im.cast(unsigned)
                assert im2.avg() == 0

        # casting very positive pixels to a signed format should clip to max
        im = (pyvips.Image.black(1, 1) + max_value["uint"]).cast("uint")
        assert im.avg() == max_value["uint"]
        im2 = im.cast("int")
        assert im2.avg() == max_value["int"]
        im = (pyvips.Image.black(1, 1) + max_value["ushort"]).cast("ushort")
        im2 = im.cast("short")
        assert im2.avg() == max_value["short"]
        im = (pyvips.Image.black(1, 1) + max_value["uchar"]).cast("uchar")
        im2 = im.cast("char")
        assert im2.avg() == max_value["char"]

    def test_band_and(self):
        def band_and(x):
            if isinstance(x, pyvips.Image):
                return x.bandand()
            else:
                return [reduce(lambda a, b: int(a) & int(b), x)]

        self.run_unary(self.all_images, band_and, fmt=int_formats)

    def test_band_or(self):
        def band_or(x):
            if isinstance(x, pyvips.Image):
                return x.bandor()
            else:
                return [reduce(lambda a, b: int(a) | int(b), x)]

        self.run_unary(self.all_images, band_or, fmt=int_formats)

    def test_band_eor(self):
        def band_eor(x):
            if isinstance(x, pyvips.Image):
                return x.bandeor()
            else:
                return [reduce(lambda a, b: int(a) ^ int(b), x)]

        self.run_unary(self.all_images, band_eor, fmt=int_formats)

    def test_bandjoin(self):
        def bandjoin(x, y):
            if isinstance(x, pyvips.Image) and isinstance(y, pyvips.Image):
                return x.bandjoin(y)
            else:
                return x + y

        self.run_binary(self.all_images, bandjoin)

    def test_bandjoin_const(self):
        x = self.colour.bandjoin(1)
        assert x.bands == 4
        assert x[3].avg() == 1

        x = self.colour.bandjoin([1, 2])
        assert x.bands == 5
        assert x[3].avg() == 1
        assert x[4].avg() == 2

    def test_addalpha(self):
        x = self.colour.addalpha()
        assert x.bands == 4
        assert x[3].avg() == 255  # self.colour is srgb

    def test_bandmean(self):
        def bandmean(x):
            if isinstance(x, pyvips.Image):
                return x.bandmean()
            else:
                return [sum(x) // len(x)]

        self.run_unary(self.all_images, bandmean, fmt=noncomplex_formats)

    def test_bandrank(self):
        def median(x, y):
            joined = [[a, b] for a, b in zip(x, y)]
            # .sort() isn't a function, so we have to run this as a separate
            # pass
            [z.sort() for z in joined]
            return [z[len(z) // 2] for z in joined]

        def bandrank(x, y):
            if isinstance(x, pyvips.Image) and isinstance(y, pyvips.Image):
                return x.bandrank([y])
            else:
                return median(x, y)

        self.run_binary(self.all_images, bandrank, fmt=noncomplex_formats)

        # we can mix images and constants, and set the index arg
        a = self.mono.bandrank([2], index=0)
        b = (self.mono < 2).ifthenelse(self.mono, 2)
        assert (a - b).abs().min() == 0

    def test_copy(self):
        x = self.colour.copy(interpretation=pyvips.Interpretation.LAB)
        assert x.interpretation == pyvips.Interpretation.LAB
        x = self.colour.copy(xres=42)
        assert x.xres == 42
        x = self.colour.copy(yres=42)
        assert x.yres == 42
        x = self.colour.copy(xoffset=42)
        assert x.xoffset == 42
        x = self.colour.copy(yoffset=42)
        assert x.yoffset == 42
        x = self.colour.copy(coding=pyvips.Coding.NONE)
        assert x.coding == pyvips.Coding.NONE

    def test_bandfold(self):
        x = self.mono.bandfold()
        assert x.width == 1
        assert x.bands == self.mono.width

        y = x.bandunfold()
        assert y.width == self.mono.width
        assert y.bands == 1
        assert x.avg() == y.avg()

        x = self.mono.bandfold(factor=2)
        assert x.width == self.mono.width / 2
        assert x.bands == 2

        y = x.bandunfold(factor=2)
        assert y.width == self.mono.width
        assert y.bands == 1
        assert x.avg() == y.avg()

    def test_byteswap(self):
        x = self.mono.cast("ushort")
        y = x.byteswap().byteswap()
        assert x.width == y.width
        assert x.height == y.height
        assert x.bands == y.bands
        assert x.avg() == y.avg()

    def test_embed(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            im = test.embed(20, 20,
                            self.colour.width + 40,
                            self.colour.height + 40)
            pixel = im(10, 10)
            assert_almost_equal_objects(pixel, [0, 0, 0])
            pixel = im(30, 30)
            assert_almost_equal_objects(pixel, [2, 3, 4])
            pixel = im(im.width - 10, im.height - 10)
            assert_almost_equal_objects(pixel, [0, 0, 0])

            im = test.embed(20, 20,
                            self.colour.width + 40,
                            self.colour.height + 40,
                            extend=pyvips.Extend.COPY)
            pixel = im(10, 10)
            assert_almost_equal_objects(pixel, [2, 3, 4])
            pixel = im(im.width - 10, im.height - 10)
            assert_almost_equal_objects(pixel, [2, 3, 4])

            im = test.embed(20, 20,
                            self.colour.width + 40,
                            self.colour.height + 40,
                            extend=pyvips.Extend.BACKGROUND,
                            background=[7, 8, 9])
            pixel = im(10, 10)
            assert_almost_equal_objects(pixel, [7, 8, 9])
            pixel = im(im.width - 10, im.height - 10)
            assert_almost_equal_objects(pixel, [7, 8, 9])

            im = test.embed(20, 20,
                            self.colour.width + 40,
                            self.colour.height + 40,
                            extend=pyvips.Extend.WHITE)
            pixel = im(10, 10)
            # uses 255 in all bytes of ints, 255.0 for float
            pixel = [int(x) & 0xff for x in pixel]
            assert_almost_equal_objects(pixel, [255, 255, 255])
            pixel = im(im.width - 10, im.height - 10)
            pixel = [int(x) & 0xff for x in pixel]
            assert_almost_equal_objects(pixel, [255, 255, 255])

    def test_gravity(self):
        im = pyvips.Image.black(1, 1) + 255

        positions = [
            ['centre', 1, 1],
            ['north', 1, 0],
            ['south', 1, 2],
            ['east', 2, 1],
            ['west', 0, 1],
            ['north-east', 2, 0],
            ['south-east', 2, 2],
            ['south-west', 0, 2],
            ['north-west', 0, 0]
        ]

        for direction, x, y in positions:
            im2 = im.gravity(direction, 3, 3)
            assert_almost_equal_objects(im2(x, y), [255])
            assert_almost_equal_objects(im2.avg(), 255.0 / 9.0)

    def test_extract(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            pixel = test(30, 30)
            assert_almost_equal_objects(pixel, [2, 3, 4])

            sub = test.extract_area(25, 25, 10, 10)

            pixel = sub(5, 5)
            assert_almost_equal_objects(pixel, [2, 3, 4])

            sub = test.extract_band(1, n=2)

            pixel = sub(30, 30)
            assert_almost_equal_objects(pixel, [3, 4])

    def test_slice(self):
        test = self.colour
        bands = [x.avg() for x in test]

        x = test[0].avg()
        assert x == bands[0]

        x = test[-1].avg()
        assert_almost_equal_objects(x, bands[2])

        x = [i.avg() for i in test[1:3]]
        assert_almost_equal_objects(x, bands[1:3])

        x = [i.avg() for i in test[1:-1]]
        assert_almost_equal_objects(x, bands[1:-1])

        x = [i.avg() for i in test[:2]]
        assert_almost_equal_objects(x, bands[:2])

        x = [i.avg() for i in test[1:]]
        assert_almost_equal_objects(x, bands[1:])

        x = [i.avg() for i in test[-1]]
        assert_almost_equal_objects(x, bands[-1])

    def test_crop(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            pixel = test(30, 30)
            assert_almost_equal_objects(pixel, [2, 3, 4])

            sub = test.crop(25, 25, 10, 10)

            pixel = sub(5, 5)
            assert_almost_equal_objects(pixel, [2, 3, 4])

    def test_smartcrop(self):
        test = self.image.smartcrop(100, 100)
        assert test.width == 100
        assert test.height == 100

    def test_smartcrop_attention(self):
        test, opts = self.image.smartcrop(
            100, 100,
            interesting=pyvips.enums.Interesting.ATTENTION,
            attention_x=True, attention_y=True)
        assert test.width == 100
        assert test.height == 100

        assert opts["attention_x"] == 199
        assert opts["attention_y"] == 234

    def test_smartcrop_rgba(self):
        rgba = pyvips.Image.new_from_file(RGBA_FILE)
        test, opts = rgba.smartcrop(
            80, 60,
            attention_x=True, attention_y=True)
        assert test.width == 80
        assert test.height == 60
        assert opts["attention_x"] == 20
        assert opts["attention_y"] == 124

    def test_smartcrop_rgba_premultiplied(self):
        rgba = pyvips.Image.new_from_file(RGBA_FILE)
        test, opts = rgba.premultiply().smartcrop(
            80, 60,
            premultiplied=True,
            attention_x=True, attention_y=True)
        assert test.width == 80
        assert test.height == 60
        assert opts["attention_x"] == 20
        assert opts["attention_y"] == 124

    def test_falsecolour(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            im = test.falsecolour()

            assert im.width == test.width
            assert im.height == test.height
            assert im.bands == 3

            pixel = im(30, 30)
            assert_almost_equal_objects(pixel, [20, 0, 41])

    def test_flatten(self):
        for fmt in unsigned_formats + [pyvips.BandFormat.SHORT,
                                       pyvips.BandFormat.INT] + float_formats:
            mx = 255
            alpha = mx / 2.0
            nalpha = mx - alpha
            test = self.colour.bandjoin(alpha).cast(fmt)
            pixel = test(30, 30)

            predict = [int(x) * alpha / mx for x in pixel[:-1]]

            im = test.flatten()

            assert im.bands == 3
            pixel = im(30, 30)
            for x, y in zip(pixel, predict):
                # we use float arithetic for int and uint, so the rounding
                # differs ... don't require huge accuracy
                assert abs(x - y) < 2

            im = test.flatten(background=[100, 100, 100])

            pixel = test(30, 30)
            predict = [int(x) * alpha / mx + (100 * nalpha) / mx
                       for x in pixel[:-1]]

            assert im.bands == 3
            pixel = im(30, 30)
            for x, y in zip(pixel, predict):
                assert abs(x - y) < 2

        # if the image has max_alpha less than the numeric range of the
        # format, we can get out of range values ... check they are clipped
        # correctly
        rgba = pyvips.Image.new_from_file(RGBA_FILE)

        im = rgba * 256
        im = im.cast("ushort")
        im = im.flatten()

        im2 = rgba * 256
        im2 = im2.flatten()
        im2 = im2.cast("ushort")

        assert(abs(im - im2).max() == 0)

    def test_premultiply(self):
        for fmt in unsigned_formats + [pyvips.BandFormat.SHORT,
                                       pyvips.BandFormat.INT] + float_formats:
            mx = 255
            alpha = mx / 2.0
            test = self.colour.bandjoin(alpha).cast(fmt)
            pixel = test(30, 30)

            predict = [int(x) * alpha / mx for x in pixel[:-1]] + [alpha]

            im = test.premultiply()

            assert im.bands == test.bands
            pixel = im(30, 30)
            for x, y in zip(pixel, predict):
                # we use float arithetic for int and uint, so the rounding
                # differs ... don't require huge accuracy
                assert abs(x - y) < 2

    def test_composite(self):
        # 50% transparent image
        overlay = self.colour.bandjoin(128)
        base = self.colour + 100
        comp = base.composite(overlay, "over")

        assert_almost_equal_objects(comp(0, 0), [51.8, 52.8, 53.8, 255],
                                    threshold=0.1)

    def test_unpremultiply(self):
        for fmt in unsigned_formats + [pyvips.BandFormat.SHORT,
                                       pyvips.BandFormat.INT] + float_formats:
            mx = 255
            alpha = mx / 2.0
            test = self.colour.bandjoin(alpha).cast(fmt)
            pixel = test(30, 30)

            predict = [int(x) / (alpha / mx) for x in pixel[:-1]] + [alpha]

            im = test.unpremultiply()

            assert im.bands == test.bands
            pixel = im(30, 30)
            for x, y in zip(pixel, predict):
                # we use float arithetic for int and uint, so the rounding
                # differs ... don't require huge accuracy
                assert abs(x - y) < 2

    def test_flip(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            result = test.fliphor()
            result = result.flipver()
            result = result.fliphor()
            result = result.flipver()

            diff = (test - result).abs().max()

            assert diff == 0

    def test_gamma(self):
        exponent = 2.4
        for fmt in noncomplex_formats:
            mx = max_value[fmt]
            test = (self.colour + mx / 2.0).cast(fmt)

            norm = mx ** exponent / mx
            result = test.gamma()
            before = test(30, 30)
            after = result(30, 30)
            predict = [x ** exponent / norm for x in before]
            for a, b in zip(after, predict):
                # ie. less than 1% error, rounding on 7-bit images
                # means this is all we can expect
                assert abs(a - b) < mx / 100.0

        exponent = 1.2
        for fmt in noncomplex_formats:
            mx = max_value[fmt]
            test = (self.colour + mx / 2.0).cast(fmt)

            norm = mx ** exponent / mx
            result = test.gamma(exponent=1.0 / 1.2)
            before = test(30, 30)
            after = result(30, 30)
            predict = [x ** exponent / norm for x in before]
            for a, b in zip(after, predict):
                # ie. less than 1% error, rounding on 7-bit images
                # means this is all we can expect
                assert abs(a - b) < mx / 100.0

    def test_grid(self):
        test = self.colour.replicate(1, 12)
        assert test.width == self.colour.width
        assert test.height == self.colour.height * 12

        for fmt in all_formats:
            im = test.cast(fmt)
            result = im.grid(test.width, 3, 4)
            assert result.width == self.colour.width * 3
            assert result.height == self.colour.height * 4

            before = im(10, 10)
            after = result(10 + test.width * 2, 10 + test.width * 2)
            assert_almost_equal_objects(before, after)

            before = im(50, 50)
            after = result(50 + test.width * 2, 50 + test.width * 2)
            assert_almost_equal_objects(before, after)

    def test_ifthenelse(self):
        test = self.mono > 3
        for x in all_formats:
            for y in all_formats:
                t = (self.colour + 10).cast(x)
                e = self.colour.cast(y)
                r = test.ifthenelse(t, e)

                assert r.width == self.colour.width
                assert r.height == self.colour.height
                assert r.bands == self.colour.bands

                predict = e(10, 10)
                result = r(10, 10)
                assert_almost_equal_objects(result, predict)

                predict = t(50, 50)
                result = r(50, 50)
                assert_almost_equal_objects(result, predict)

        test = self.colour > 3
        for x in all_formats:
            for y in all_formats:
                t = (self.mono + 10).cast(x)
                e = self.mono.cast(y)
                r = test.ifthenelse(t, e)

                assert r.width == self.colour.width
                assert r.height == self.colour.height
                assert r.bands == self.colour.bands

                cp = test(10, 10)
                tp = t(10, 10) * 3
                ep = e(10, 10) * 3
                predict = [te if ce != 0 else ee
                           for ce, te, ee in zip(cp, tp, ep)]
                result = r(10, 10)
                assert_almost_equal_objects(result, predict)

                cp = test(50, 50)
                tp = t(50, 50) * 3
                ep = e(50, 50) * 3
                predict = [te if ce != 0 else ee
                           for ce, te, ee in zip(cp, tp, ep)]
                result = r(50, 50)
                assert_almost_equal_objects(result, predict)

        test = self.colour > 3
        for x in all_formats:
            for y in all_formats:
                t = (self.mono + 10).cast(x)
                e = self.mono.cast(y)
                r = test.ifthenelse(t, e, blend=True)

                assert r.width == self.colour.width
                assert r.height == self.colour.height
                assert r.bands == self.colour.bands

                result = r(10, 10)
                assert_almost_equal_objects(result, [3, 3, 13])

        test = self.mono > 3
        r = test.ifthenelse([1, 2, 3], self.colour)
        assert r.width == self.colour.width
        assert r.height == self.colour.height
        assert r.bands == self.colour.bands
        assert r.format == self.colour.format
        assert r.interpretation == self.colour.interpretation
        result = r(10, 10)
        assert_almost_equal_objects(result, [2, 3, 4])
        result = r(50, 50)
        assert_almost_equal_objects(result, [1, 2, 3])

        test = self.mono
        r = test.ifthenelse([1, 2, 3], self.colour, blend=True)
        assert r.width == self.colour.width
        assert r.height == self.colour.height
        assert r.bands == self.colour.bands
        assert r.format == self.colour.format
        assert r.interpretation == self.colour.interpretation
        result = r(10, 10)
        assert_almost_equal_objects(result, [2, 3, 4], threshold=0.1)
        result = r(50, 50)
        assert_almost_equal_objects(result, [3.0, 4.9, 6.9], threshold=0.1)

    def test_switch(self):
        x = pyvips.Image.grey(256, 256, uchar=True)

        # slice into two at 128, we should get 50% of pixels in each half
        index = pyvips.Image.switch([x < 128, x >= 128])
        assert index.avg() == 0.5

        # slice into four
        index = pyvips.Image.switch([
            x < 64,
            x >= 64 and x < 128,
            x >= 128 and x < 192,
            x >= 192
        ])
        assert index.avg() == 1.5

        # no match should return n + 1
        index = pyvips.Image.switch([x == 1000, x == 2000])
        assert index.avg() == 2

    def test_insert(self):
        for x in all_formats:
            for y in all_formats:
                main = self.mono.cast(x)
                sub = self.colour.cast(y)
                r = main.insert(sub, 10, 10)

                assert r.width == main.width
                assert r.height == main.height
                assert r.bands == sub.bands

                a = r(10, 10)
                b = sub(0, 0)
                assert_almost_equal_objects(a, b)

                a = r(0, 0)
                b = main(0, 0) * 3
                assert_almost_equal_objects(a, b)

        for x in all_formats:
            for y in all_formats:
                main = self.mono.cast(x)
                sub = self.colour.cast(y)
                r = main.insert(sub, 10, 10, expand=True, background=100)

                assert r.width == main.width + 10
                assert r.height == main.height + 10
                assert r.bands == sub.bands

                a = r(r.width - 5, 5)
                assert_almost_equal_objects(a, [100, 100, 100])

    def test_arrayjoin(self):
        max_width = 0
        max_height = 0
        max_bands = 0
        for image in self.all_images:
            if image.width > max_width:
                max_width = image.width
            if image.height > max_height:
                max_height = image.height
            if image.bands > max_bands:
                max_bands = image.bands

        im = pyvips.Image.arrayjoin(self.all_images)
        assert im.width == max_width * len(self.all_images)
        assert im.height == max_height
        assert im.bands == max_bands

        im = pyvips.Image.arrayjoin(self.all_images, across=1)
        assert im.width == max_width
        assert im.height == max_height * len(self.all_images)
        assert im.bands == max_bands

        im = pyvips.Image.arrayjoin(self.all_images, shim=10)
        assert im.width == max_width * len(self.all_images) + 10 * (len(self.all_images) - 1)  # noqa: E501
        assert im.height == max_height
        assert im.bands == max_bands

    def test_msb(self):
        for fmt in unsigned_formats:
            mx = max_value[fmt]
            size = sizeof_format[fmt]
            test = (self.colour + mx / 8.0).cast(fmt)
            im = test.msb()

            before = test(10, 10)
            predict = [int(x) >> ((size - 1) * 8) for x in before]
            result = im(10, 10)
            assert_almost_equal_objects(result, predict)

            before = test(50, 50)
            predict = [int(x) >> ((size - 1) * 8) for x in before]
            result = im(50, 50)
            assert_almost_equal_objects(result, predict)

        for fmt in signed_formats:
            mx = max_value[fmt]
            size = sizeof_format[fmt]
            test = (self.colour + mx / 8.0).cast(fmt)
            im = test.msb()

            before = test(10, 10)
            predict = [128 + (int(x) >> ((size - 1) * 8)) for x in before]
            result = im(10, 10)
            assert_almost_equal_objects(result, predict)

            before = test(50, 50)
            predict = [128 + (int(x) >> ((size - 1) * 8)) for x in before]
            result = im(50, 50)
            assert_almost_equal_objects(result, predict)

        for fmt in unsigned_formats:
            mx = max_value[fmt]
            size = sizeof_format[fmt]
            test = (self.colour + mx / 8.0).cast(fmt)
            im = test.msb(band=1)

            before = [test(10, 10)[1]]
            predict = [int(x) >> ((size - 1) * 8) for x in before]
            result = im(10, 10)
            assert_almost_equal_objects(result, predict)

            before = [test(50, 50)[1]]
            predict = [int(x) >> ((size - 1) * 8) for x in before]
            result = im(50, 50)
            assert_almost_equal_objects(result, predict)

    def test_recomb(self):
        array = [[0.2, 0.5, 0.3]]

        def recomb(x):
            if isinstance(x, pyvips.Image):
                return x.recomb(array)
            else:
                sum = 0
                for i, c in zip(array[0], x):
                    sum += i * c
                return [sum]

        self.run_unary([self.colour], recomb, fmt=noncomplex_formats)

    def test_replicate(self):
        for fmt in all_formats:
            im = self.colour.cast(fmt)

            test = im.replicate(10, 10)
            assert test.width == self.colour.width * 10
            assert test.height == self.colour.height * 10

            before = im(10, 10)
            after = test(10 + im.width * 2, 10 + im.width * 2)
            assert_almost_equal_objects(before, after)

            before = im(50, 50)
            after = test(50 + im.width * 2, 50 + im.width * 2)
            assert_almost_equal_objects(before, after)

    def test_rot45(self):
        # test has a quarter-circle in the bottom right
        test = self.colour.crop(0, 0, 51, 51)
        for fmt in all_formats:
            im = test.cast(fmt)

            im2 = im.rot45()
            before = im(50, 50)
            after = im2(25, 50)
            assert_almost_equal_objects(before, after)

            for a, b in zip(rot45_angles, rot45_angle_bonds):
                im2 = im.rot45(angle=a)
                after = im2.rot45(angle=b)
                diff = (after - im).abs().max()
                assert diff == 0

    def test_rot(self):
        # test has a quarter-circle in the bottom right
        test = self.colour.crop(0, 0, 51, 51)
        for fmt in all_formats:
            im = test.cast(fmt)

            im2 = im.rot(pyvips.Angle.D90)
            before = im(50, 50)
            after = im2(0, 50)
            assert_almost_equal_objects(before, after)

            for a, b in zip(rot_angles, rot_angle_bonds):
                im2 = im.rot(a)
                after = im2.rot(b)
                diff = (after - im).abs().max()
                assert diff == 0

    def test_autorot(self):
        rotation_images = os.path.join(IMAGES, 'rotation')
        files = os.listdir(rotation_images)
        files.sort()

        meta = {
            0: {'w': 290, 'h': 442},
            1: {'w': 308, 'h': 410},
            2: {'w': 308, 'h': 410},
            3: {'w': 308, 'h': 410},
            4: {'w': 308, 'h': 410},
            5: {'w': 231, 'h': 308},
            6: {'w': 231, 'h': 308},
            7: {'w': 231, 'h': 308},
            8: {'w': 231, 'h': 308},
        }

        i = 0
        for f in files:
            if '.autorot.' not in f and not f.startswith('.'):
                source_filename = os.path.join(rotation_images, f)

                actual_filename = temp_filename(self.tempdir, '.jpg')

                pyvips.Image.new_from_file(source_filename).autorot().write_to_file(actual_filename)

                actual = pyvips.Image.new_from_file(actual_filename)

                assert actual.width == meta[i]['w']
                assert actual.height == meta[i]['h']
                assert actual.get('orientation') if actual.get_typeof('orientation') else None is None
                i = i + 1

    def test_scaleimage(self):
        for fmt in noncomplex_formats:
            test = self.colour.cast(fmt)

            im = test.scaleimage()
            assert im.max() == 255
            assert im.min() == 0

            im = test.scaleimage(log=True)
            assert im.max() == 255

    def test_subsample(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            im = test.subsample(3, 3)
            assert im.width == test.width // 3
            assert im.height == test.height // 3

            before = test(60, 60)
            after = im(20, 20)
            assert_almost_equal_objects(before, after)

    def test_zoom(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            im = test.zoom(3, 3)
            assert im.width == test.width * 3
            assert im.height == test.height * 3

            before = test(50, 50)
            after = im(150, 150)
            assert_almost_equal_objects(before, after)

    def test_wrap(self):
        for fmt in all_formats:
            test = self.colour.cast(fmt)

            im = test.wrap()
            assert im.width == test.width
            assert im.height == test.height

            before = test(0, 0)
            after = im(50, 50)
            assert_almost_equal_objects(before, after)

            before = test(50, 50)
            after = im(0, 0)
            assert_almost_equal_objects(before, after)


if __name__ == '__main__':
    pytest.main()



# vim: set fileencoding=utf-8 :
import pytest

import pyvips


class TestDraw:
    def test_draw_circle(self):
        im = pyvips.Image.black(100, 100)
        im = im.draw_circle(100, 50, 50, 25)
        pixel = im(25, 50)
        assert len(pixel) == 1
        assert pixel[0] == 100
        pixel = im(26, 50)
        assert len(pixel) == 1
        assert pixel[0] == 0

        im = pyvips.Image.black(100, 100)
        im = im.draw_circle(100, 50, 50, 25, fill=True)
        pixel = im(25, 50)
        assert len(pixel) == 1
        assert pixel[0] == 100
        pixel = im(26, 50)
        assert pixel[0] == 100
        pixel = im(24, 50)
        assert pixel[0] == 0

    def test_draw_flood(self):
        im = pyvips.Image.black(100, 100)
        im = im.draw_circle(100, 50, 50, 25)
        im = im.draw_flood(100, 50, 50)

        im2 = pyvips.Image.black(100, 100)
        im2 = im2.draw_circle(100, 50, 50, 25, fill=True)

        diff = (im - im2).abs().max()
        assert diff == 0

    def test_draw_flood_out_of_bounds(self):
        im = pyvips.Image.black(100, 100)
        with pytest.raises(pyvips.error.Error):
            im.draw_flood(100, 200, 50)
        with pytest.raises(pyvips.error.Error):
            im.draw_flood(100, 50, 200)
        with pytest.raises(pyvips.error.Error):
            im.draw_flood(100, -1, 50)
        with pytest.raises(pyvips.error.Error):
            im.draw_flood(100, 50, -1)

    def test_draw_image(self):
        im = pyvips.Image.black(51, 51)
        im = im.draw_circle(100, 25, 25, 25, fill=True)

        im2 = pyvips.Image.black(100, 100)
        im2 = im2.draw_image(im, 25, 25)

        im3 = pyvips.Image.black(100, 100)
        im3 = im3.draw_circle(100, 50, 50, 25, fill=True)

        diff = (im2 - im3).abs().max()
        assert diff == 0

    def test_draw_line(self):
        im = pyvips.Image.black(100, 100)
        im = im.draw_line(100, 0, 0, 100, 0)
        pixel = im(0, 0)
        assert len(pixel) == 1
        assert pixel[0] == 100
        pixel = im(0, 1)
        assert len(pixel) == 1
        assert pixel[0] == 0

    def test_draw_mask(self):
        mask = pyvips.Image.black(51, 51)
        mask = mask.draw_circle(128, 25, 25, 25, fill=True)

        im = pyvips.Image.black(100, 100)
        im = im.draw_mask(200, mask, 25, 25)

        im2 = pyvips.Image.black(100, 100)
        im2 = im2.draw_circle(100, 50, 50, 25, fill=True)

        diff = (im - im2).abs().max()
        assert diff == 0

    def test_draw_rect(self):
        im = pyvips.Image.black(100, 100)
        im = im.draw_rect(100, 25, 25, 50, 50, fill=True)

        im2 = pyvips.Image.black(100, 100)
        for y in range(25, 75):
            im2 = im2.draw_line(100, 25, y, 74, y)

        diff = (im - im2).abs().max()
        assert diff == 0

    def test_draw_smudge(self):
        im = pyvips.Image.black(100, 100)
        im = im.draw_circle(100, 50, 50, 25, fill=True)

        im2 = im.draw_smudge(10, 10, 50, 50)

        im3 = im.crop(10, 10, 50, 50)

        im4 = im2.draw_image(im3, 10, 10)

        diff = (im4 - im).abs().max()
        assert diff == 0


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python3

# walk vips and generate a list of all operators and their descriptions
# for docs

# this needs pyvips
#   pip install --user pyvips

# rebuild with:
#   ./gen-function-list.py > function-list.md

# sample output:
# | `gamma` | Gamma an image | [method@Image.gamma] |

from pyvips import Introspect, Operation, Error, \
    ffi, type_map, type_from_name, nickname_find

# for VipsOperationFlags
_OPERATION_DEPRECATED = 8


def gen_function(operation_name, overloads):
    intro = Introspect.get(operation_name)

    # Keep-in-sync with fixup-gir-for-doc.py
    image_funcs = [
        'arrayjoin',
        'bandjoin',
        'bandrank',
        'composite',
        'sum',
        'switch',
    ]

    if operation_name in image_funcs:
        c_operations = f'[func@Image.{operation_name}]'
    else:
        operation_type = 'ctor' if intro.member_x is None else 'method'
        operation_class = 'Blob' if operation_name == 'profile_load' else 'Image'
        c_operations = f'[{operation_type}@{operation_class}.{operation_name}]'

    if overloads:
        c_operations += ', ' + (', '.join(f'[method@Image.{n}]' for n in overloads))

    description = intro.description.capitalize()
    # hide library-specific implementation details
    description = (description
                   .replace('imagemagick7', 'imagemagick')
                   .replace(' (pdfium)', '')
                   .replace(' (poppler)', ''))
    result = f"| `{operation_name}` | {description} | {c_operations} |"

    return result


def gen_function_list():
    all_nicknames = []

    def add_nickname(gtype, a, b):
        nickname = nickname_find(gtype)
        try:
            # can fail for abstract types
            intro = Introspect.get(nickname)

            # we are only interested in non-deprecated operations
            if (intro.flags & _OPERATION_DEPRECATED) == 0:
                all_nicknames.append(nickname)
        except Error:
            pass

        type_map(gtype, add_nickname)

        return ffi.NULL

    type_map(type_from_name('VipsOperation'), add_nickname)

    # make list unique and sort
    all_nicknames = list(set(all_nicknames))
    all_nicknames.sort()

    # make dict with overloads
    overloads = {
        'bandbool': ['bandand', 'bandor', 'bandeor', 'bandmean'],
        'bandjoin': ['bandjoin2'],
        'bandjoin_const': ['bandjoin_const1'],
        'boolean': ['andimage', 'orimage', 'eorimage', 'lshift', 'rshift'],
        'cast': ['cast_uchar', 'cast_char', 'cast_ushort', 'cast_short',
            'cast_uint', 'cast_int', 'cast_float', 'cast_double', 
            'cast_complex', 'cast_dpcomplex'],
        'complex': ['polar', 'rect', 'conj'],
        'complex2': ['cross_phase'],
        'complexget': ['real', 'imag'],
        'draw_circle': ['draw_circle1'],
        'draw_flood': ['draw_flood1'],
        'draw_line': ['draw_line1'],
        'draw_mask': ['draw_mask1'],
        'draw_rect': ['draw_rect1', 'draw_point', 'draw_point1'],
        'extract_area': ['crop'],
        'linear': ['linear1'],
        'math': ['sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'sinh', 
            'cosh', 'tanh', 'asinh', 'acosh', 'atanh', 'exp', 'exp10', 
            'log', 'log10'],
        'math2': ['pow', 'wop', 'atan2'],
        'rank': ['median'],
        'relational': ['equal', 'notequal', 'less', 'lesseq', 'more', 'moreeq'],
        'remainder_const': ['remainder_const1'],
        'round': ['floor', 'ceil', 'rint'],
    }

    overloads['boolean_const'] = \
            [o + '_const' for o in overloads['boolean']] \
            + ['boolean_const1'] \
            + [o + '_const1' for o in overloads['boolean']]

    overloads['math2_const'] = \
            [o + '_const' for o in overloads['boolean']] \
            + ['math2_const1'] \
            + [o + '_const1' for o in overloads['boolean']]

    overloads['relational_const'] = \
            [o + '_const' for o in overloads['relational']] \
            + ['relational_const1'] \
            + [o + '_const1' for o in overloads['relational']]

    for nickname in all_nicknames:
        result = gen_function(nickname, 
                overloads[nickname] if nickname in overloads else None)
        print(result)


if __name__ == '__main__':
    print("""Title: Operator index / Alphabetical

libvips has a set of operators, each of which computes some useful image
processing operation. Each operator is implemented as a [class@GObject.Object]
class, for example `VipsGamma`. Classes are identified by their unique
[property@VipsObject:nickname], in this case `gamma`.

From the command-line, C++ and most language bindings, you use the nickname
to call the operator. For example in C++:

```c++
vips::VImage fred = ...;
vips::VImage jim = fred.gamma();
```

or Python:

```python
fred = jim.gamma()
```

libvips has a set of C wrapper functions for calling operators, in this
case [method@Image.gamma]:

```c
VipsImage *fred = ...;
VipsImage *jim;

if (vips_gamma(fred, &jim, NULL))
    ...error;
```

Some operators have many C convenience functions.

# All libvips operators

This table lists all the libvips operators with their C convenience functions
and a short description. It's supposed to be useful for searching. See the
API docs each function links to for more details.

| Operator | Description | C functions |
| -------- | ----------- | ----------- |""")

    gen_function_list()



# vim: set fileencoding=utf-8 :
import pytest

import pyvips
from helpers import *


# Run a function expecting a complex image on a two-band image
def run_cmplx(fn, image):
    if image.format == pyvips.BandFormat.FLOAT:
        new_format = pyvips.BandFormat.COMPLEX
    elif image.format == pyvips.BandFormat.DOUBLE:
        new_format = pyvips.BandFormat.DPCOMPLEX
    else:
        raise pyvips.Error("run_cmplx: not float or double")

    # tag as complex, run, revert tagging
    cmplx = image.copy(bands=1, format=new_format)
    cmplx_result = fn(cmplx)

    return cmplx_result.copy(bands=2, format=image.format)


def to_polar(image):
    """Transform image coordinates to polar.

    The image is transformed so that it is wrapped around a point in the
    centre. Vertical straight lines become circles or segments of circles,
    horizontal straight lines become radial spokes.
    """
    # xy image, zero in the centre, scaled to fit image to a circle
    xy = pyvips.Image.xyz(image.width, image.height)
    xy -= [image.width / 2.0, image.height / 2.0]
    scale = min(image.width, image.height) / float(image.width)
    xy *= 2.0 / scale

    # to polar, scale vertical axis to 360 degrees
    index = run_cmplx(lambda x: x.polar(), xy)
    index *= [1, image.height / 360.0]

    return image.mapim(index)


def to_rectangular(image):
    """Transform image coordinates to rectangular.

    The image is transformed so that it is unwrapped from a point in the
    centre. Circles or segments of circles become vertical straight lines,
    radial lines become horizontal lines.
    """
    # xy image, vertical scaled to 360 degrees
    xy = pyvips.Image.xyz(image.width, image.height)
    xy *= [1, 360.0 / image.height]

    # to rect, scale to image rect
    index = run_cmplx(lambda x: x.rect(), xy)
    scale = min(image.width, image.height) / float(image.width)
    index *= scale / 2.0
    index += [image.width / 2.0, image.height / 2.0]

    return image.mapim(index)


class TestResample:
    def test_affine(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)

        # vsqbs is non-interpolatory, don't test this way
        for name in ["nearest", "bicubic", "bilinear", "nohalo", "lbb"]:
            x = im
            interpolate = pyvips.Interpolate.new(name)
            for i in range(4):
                x = x.affine([0, 1, 1, 0], interpolate=interpolate)

            assert (x - im).abs().max() == 0

    def test_reduce(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)
        # cast down to 0-127, the smallest range, so we aren't messed up by
        # clipping
        im = im.cast(pyvips.BandFormat.CHAR)

        for fac in [1, 1.1, 1.5, 1.999]:
            for fmt in all_formats:
                for kernel in ["nearest", "linear",
                               "cubic", "lanczos2",
                               "lanczos3", "mks2013", "mks2021"]:
                    x = im.cast(fmt)
                    r = x.reduce(fac, fac, kernel=kernel)
                    d = abs(r.avg() - im.avg())
                    assert d < 2

        # try constant images ... should not change the constant
        for const in [0, 1, 2, 254, 255]:
            im = (pyvips.Image.black(10, 10) + const).cast("uchar")
            for kernel in ["nearest", "linear",
                           "cubic", "lanczos2",
                           "lanczos3", "mks2013", "mks2021"]:
                # print "testing kernel =", kernel
                # print "testing const =", const
                shr = im.reduce(2, 2, kernel=kernel)
                d = abs(shr.avg() - im.avg())
                assert d == 0

        # https://github.com/libvips/libvips/issues/4864
        if have("ppmload"):
            im = pyvips.Image.new_from_buffer(b'P6\n2 2\n255\n'
                                              b'\xff\x00\x00' b'\x00\xff\x00'
                                              b'\x00\x00\xff' b'\xff\xff\x00', "")
            im2 = im.reduceh(1.5, kernel="nearest")
            assert im2.width == 1

    def test_resize(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)
        im2 = im.resize(0.25)
        # in py3, round() does not round to nearest in the obvious way, so we
        # have to do it by hand
        assert im2.width == int(im.width / 4.0 + 0.5)
        assert im2.height == int(im.height / 4.0 + 0.5)

        # test geometry rounding corner case
        im = pyvips.Image.black(100, 1)
        x = im.resize(0.5)
        assert x.width == 50
        assert x.height == 1

        # test whether we use double-precision calculations in reduce{h,v}
        im = pyvips.Image.black(1600, 1000)
        x = im.resize(10.0 / im.width)
        assert x.width == 10
        assert x.height == 6

        # test round-up option of shrink
        im = pyvips.Image.black(2049 - 2, 2047 - 2, bands=3)
        im = im.embed(1, 1, 2049, 2047,
                      extend=pyvips.Extend.BACKGROUND,
                      background=[255, 0, 0])
        for scale in [8, 9.4, 16]:
            x = im.resize(1 / scale, vscale=1 / scale)

            for point in ([(round(x.width / 2), 0),
                           (x.width - 1, round(x.height / 2)),
                           (round(x.width / 2), x.height - 1),
                           (0, round(x.height / 2))]):
                y = x(*point)[0]
                assert y != 0

    def test_shrink(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)
        im2 = im.shrink(4, 4)
        # in py3, round() does not round to nearest in the obvious way, so we
        # have to do it by hand
        assert im2.width == int(im.width / 4.0 + 0.5)
        assert im2.height == int(im.height / 4.0 + 0.5)
        assert abs(im.avg() - im2.avg()) < 1

        im2 = im.shrink(2.5, 2.5)
        assert im2.width == int(im.width / 2.5 + 0.5)
        assert im2.height == int(im.height / 2.5 + 0.5)
        assert abs(im.avg() - im2.avg()) < 1

        # https://github.com/libvips/libvips/issues/4864
        if have("ppmload"):
            im = pyvips.Image.new_from_buffer(b'P6\n2 2\n255\n'
                                              b'\xff\x00\x00' b'\x00\xff\x00'
                                              b'\x00\x00\xff' b'\xff\xff\x00', "")
            im2 = im.shrinkh(2)
            assert im2.width == 1
            assert abs(im.avg() - im2.avg()) < 1

    def test_thumbnail(self):
        im = pyvips.Image.thumbnail(JPEG_FILE, 100)

        assert im.height == 100
        assert im.bands == 3
        assert im.bands == 3

        # the average shouldn't move too much
        im_orig = pyvips.Image.new_from_file(JPEG_FILE)
        assert abs(im_orig.avg() - im.avg()) < 1

        # make sure we always get the right width
        for height in range(440, 1, -13):
            im = pyvips.Image.thumbnail(JPEG_FILE, height)
            assert im.height == height

        # should fit one of width or height
        im = pyvips.Image.thumbnail(JPEG_FILE, 100, height=300)
        assert im.width == 100
        assert im.height != 300
        im = pyvips.Image.thumbnail(JPEG_FILE, 300, height=100)
        assert im.width != 300
        assert im.height == 100

        # with @crop, should fit both width and height
        im = pyvips.Image.thumbnail(JPEG_FILE, 100,
                                    height=300, crop=True)
        assert im.width == 100
        assert im.height == 300

        im1 = pyvips.Image.thumbnail(JPEG_FILE, 100)
        with open(JPEG_FILE, 'rb') as f:
            buf = f.read()
        im2 = pyvips.Image.thumbnail_buffer(buf, 100)
        assert abs(im1.avg() - im2.avg()) < 1

        # should be able to thumbnail many-page tiff
        im = pyvips.Image.thumbnail(OME_FILE, 100)
        assert im.width == 100
        assert im.height == 38

        # should be able to thumbnail individual pages from many-page tiff
        im1 = pyvips.Image.thumbnail(OME_FILE + "[page=0]", 100)
        assert im1.width == 100
        assert im1.height == 38
        im2 = pyvips.Image.thumbnail(OME_FILE + "[page=1]", 100)
        assert im2.width == 100
        assert im2.height == 38
        assert (im1 - im2).abs().max() != 0

        # should be able to thumbnail entire many-page tiff as a toilet-roll
        # image
        im = pyvips.Image.thumbnail(OME_FILE + "[n=-1]", 100)
        assert im.width == 100
        assert im.height == 570

        # should be able to thumbnail a single-page tiff in a buffer
        im1 = pyvips.Image.thumbnail(TIF_FILE, 100)
        with open(TIF_FILE, 'rb') as f:
            buf = f.read()
        im2 = pyvips.Image.thumbnail_buffer(buf, 100)
        assert abs(im1.avg() - im2.avg()) < 1

        # linear shrink should work on rgba images
        if have("ppmload"):
            im1 = pyvips.Image.thumbnail(RGBA_FILE, 64, linear=True)
            im2 = pyvips.Image.new_from_file(RGBA_CORRECT_FILE)
            assert abs(im1.flatten(background=255).avg() - im2.avg()) < 1

        # thumbnailing a 16-bit image should always make an 8-bit image
        rgb16_buffer = pyvips.Image \
                .new_from_file(JPEG_FILE) \
                .colourspace("rgb16") \
                .write_to_buffer(".png")
        thumb = pyvips.Image.thumbnail_buffer(rgb16_buffer, 128)
        assert thumb.format == "uchar"

        if have("heifload"):
            # this image is orientation 6 ... thumbnail should flip it
            im = pyvips.Image.new_from_file(AVIF_FILE)
            thumb = pyvips.Image.thumbnail(AVIF_FILE, 100)

            # thumb should be portrait
            assert thumb.width < thumb.height
            assert thumb.height == 100

    def test_thumbnail_icc(self):
        im = pyvips.Image.thumbnail(JPEG_FILE_XYB, 442, output_profile="srgb")

        assert im.width == 290
        assert im.height == 442
        assert im.bands == 3

        # the colour distance should not deviate too much
        # (i.e. the embedded profile should not be ignored)
        im_orig = pyvips.Image.new_from_file(JPEG_FILE)
        assert im_orig.de00(im).max() < 10

    # this has caused a few bugs in the past ,,,
    def test_thumbnail_uhdr_linear(self):
        im = pyvips.Image.thumbnail(UHDR_FILE, 128, linear=True)

        assert im.width == 128
        assert im.bands == 3

    def test_similarity(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)
        im2 = im.similarity(angle=90)
        im3 = im.affine([0, -1, 1, 0])
        # rounding in calculating the affine transform from the angle stops
        # this being exactly true
        assert (im2 - im3).abs().max() < 50

    def test_similarity_scale(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)
        im2 = im.similarity(scale=2)
        im3 = im.affine([2, 0, 0, 2])
        assert (im2 - im3).abs().max() == 0

    # added in 8.7
    def test_rotate(self):
        if have("rotate"):
            im = pyvips.Image.new_from_file(JPEG_FILE)
            im2 = im.rotate(90)
            im3 = im.affine([0, -1, 1, 0])
            # rounding in calculating the affine transform from the angle stops
            # this being exactly true
            assert (im2 - im3).abs().max() < 50

    def test_mapim(self):
        im = pyvips.Image.new_from_file(JPEG_FILE)

        p = to_polar(im)
        r = to_rectangular(p)

        # the left edge (which is squashed to the origin) will be badly
        # distorted, but the rest should not be too bad
        a = r.crop(50, 0, im.width - 50, im.height).gaussblur(2)
        b = im.crop(50, 0, im.width - 50, im.height).gaussblur(2)
        assert (a - b).abs().max() < 50

        # this was a bug at one point, strangely, if executed with debug
        # enabled
        mp = pyvips.Image.xyz(im.width, im.height)
        interp = pyvips.Interpolate.new('bicubic')
        assert im.mapim(mp, interpolate=interp).avg() == im.avg()


if __name__ == '__main__':
    pytest.main()



# vim: set fileencoding=utf-8 :

import math
import pytest

import pyvips
from helpers import *


class TestArithmetic:
    def run_arith(self, fn, fmt=all_formats):
        [run_image2('%s image %s %s %s' % (fn.__name__, x, y, z),
                    x.cast(y), x.cast(z), fn)
         for x in self.all_images for y in fmt for z in fmt]

    def run_arith_const(self, fn, fmt=all_formats):
        [run_const('%s scalar %s %s' % (fn.__name__, x, y),
                   fn, x.cast(y), 2)
         for x in self.all_images for y in fmt]
        [run_const('%s vector %s' % (fn.__name__, y),
                   fn, self.colour.cast(y), [1, 2, 3])
         for y in fmt]

    # run a function on an image,
    # 50,50 and 10,10 should have different values on the test image
    def run_imageunary(self, message, im, fn):
        run_cmp(message, im, 50, 50, lambda x: run_fn(fn, x))
        run_cmp(message, im, 10, 10, lambda x: run_fn(fn, x))

    def run_unary(self, images, fn, fmt=all_formats):
        [self.run_imageunary(fn.__name__ + ' image', x.cast(y), fn)
         for x in images for y in fmt]

    # run a function on a pair of images
    # 50,50 and 10,10 should have different values on the test image
    # don't loop over band elements
    def run_imagebinary(self, message, left, right, fn):
        run_cmp2(message, left, right, 50, 50, fn)
        run_cmp2(message, left, right, 10, 10, fn)

    def run_binary(self, images, fn, fmt=all_formats):
        [self.run_imagebinary(f'{fn.__name__ } {y} {x}',
                              x.cast(y), x.cast(z), fn)
         for x in images for y in fmt for z in fmt]

    def versiontuple(version_string):
        return tuple(map(int, (version_string.split('.'))))

    @classmethod
    def setup_class(cls):
        im = pyvips.Image.mask_ideal(100, 100, 0.5,
                                     reject=True, optical=True)
        cls.colour = im * [1, 2, 3] + [2, 3, 4]
        cls.mono = cls.colour.extract_band(1)
        cls.all_images = [cls.mono, cls.colour]

    @classmethod
    def teardown_class(cls):
        cls.colour = None
        cls.mono = None
        cls.all_images = None

    # test all operator overloads we define

    def test_add(self):
        def add(x, y):
            return x + y

        self.run_arith_const(add)
        self.run_arith(add)

    def test_sub(self):
        def sub(x, y):
            return x - y

        self.run_arith_const(sub)
        self.run_arith(sub)

    def test_mul(self):
        def mul(x, y):
            return x * y

        self.run_arith_const(mul)
        self.run_arith(mul)

    def test_div(self):
        def div(x, y):
            return x / y

        # (const / image) needs (image ** -1), which won't work for complex
        self.run_arith_const(div, fmt=noncomplex_formats)
        self.run_arith(div)

    def test_floordiv(self):
        def my_floordiv(x, y):
            return x // y

        # (const // image) needs (image ** -1), which won't work for complex
        self.run_arith_const(my_floordiv, fmt=noncomplex_formats)
        self.run_arith(my_floordiv, fmt=noncomplex_formats)

    def test_pow(self):
        def my_pow(x, y):
            return x ** y

        # (image ** x) won't work for complex images ... just test non-complex
        self.run_arith_const(my_pow, fmt=noncomplex_formats)
        self.run_arith(my_pow, fmt=noncomplex_formats)

    def test_and(self):
        def my_and(x, y):
            # python doesn't allow bools on float
            if isinstance(x, float):
                x = int(x)
            if isinstance(y, float):
                y = int(y)
            return x & y

        self.run_arith_const(my_and, fmt=noncomplex_formats)
        self.run_arith(my_and, fmt=noncomplex_formats)

    def test_or(self):
        def my_or(x, y):
            # python doesn't allow bools on float
            if isinstance(x, float):
                x = int(x)
            if isinstance(y, float):
                y = int(y)
            return x | y

        self.run_arith_const(my_or, fmt=noncomplex_formats)
        self.run_arith(my_or, fmt=noncomplex_formats)

    def test_xor(self):
        def my_xor(x, y):
            # python doesn't allow bools on float
            if isinstance(x, float):
                x = int(x)
            if isinstance(y, float):
                y = int(y)
            return x ^ y

        self.run_arith_const(my_xor, fmt=noncomplex_formats)
        self.run_arith(my_xor, fmt=noncomplex_formats)

    def test_more(self):
        def more(x, y):
            if isinstance(x, pyvips.Image) or isinstance(y, pyvips.Image):
                return x > y
            else:
                if x > y:
                    return 255
                else:
                    return 0

        self.run_arith_const(more)
        self.run_arith(more)

    def test_moreeq(self):
        def moreeq(x, y):
            if isinstance(x, pyvips.Image) or isinstance(y, pyvips.Image):
                return x >= y
            else:
                if x >= y:
                    return 255
                else:
                    return 0

        self.run_arith_const(moreeq)
        self.run_arith(moreeq)

    def test_less(self):
        def less(x, y):
            if isinstance(x, pyvips.Image) or isinstance(y, pyvips.Image):
                return x < y
            else:
                if x < y:
                    return 255
                else:
                    return 0

        self.run_arith_const(less)
        self.run_arith(less)

    def test_lesseq(self):
        def lesseq(x, y):
            if isinstance(x, pyvips.Image) or isinstance(y, pyvips.Image):
                return x <= y
            else:
                if x <= y:
                    return 255
                else:
                    return 0

        self.run_arith_const(lesseq)
        self.run_arith(lesseq)

    def test_equal(self):
        def equal(x, y):
            if isinstance(x, pyvips.Image) or isinstance(y, pyvips.Image):
                return x == y
            else:
                if x == y:
                    return 255
                else:
                    return 0

        self.run_arith_const(equal)
        self.run_arith(equal)

    def test_noteq(self):
        def noteq(x, y):
            if isinstance(x, pyvips.Image) or isinstance(y, pyvips.Image):
                return x != y
            else:
                if x != y:
                    return 255
                else:
                    return 0

        self.run_arith_const(noteq)
        self.run_arith(noteq)

        # comparisons against out of range values should always fail, and
        # comparisons to fractional values should always fail
        x = pyvips.Image.grey(256, 256, uchar=True)
        assert (x == 1000).max() == 0
        assert (x == 12).max() == 255
        assert (x == 12.5).max() == 0

    def test_abs(self):
        def my_abs(x):
            return abs(x)

        im = -self.colour
        self.run_unary([im], my_abs)

    def test_lshift(self):
        def my_lshift(x):
            # python doesn't allow float << int
            if isinstance(x, float):
                x = int(x)
            return x << 2

        # we don't support constant << image, treat as a unary
        self.run_unary(self.all_images, my_lshift, fmt=noncomplex_formats)

    def test_rshift(self):
        def my_rshift(x):
            # python doesn't allow float >> int
            if isinstance(x, float):
                x = int(x)
            return x >> 2

        # we don't support constant >> image, treat as a unary
        self.run_unary(self.all_images, my_rshift, fmt=noncomplex_formats)

    def test_mod(self):
        def my_mod(x):
            return x % 2

        # we don't support constant % image, treat as a unary
        self.run_unary(self.all_images, my_mod, fmt=noncomplex_formats)

    def test_pos(self):
        def my_pos(x):
            return +x

        self.run_unary(self.all_images, my_pos)

    def test_neg(self):
        def my_neg(x):
            return -x

        self.run_unary(self.all_images, my_neg)

    def test_invert(self):
        def my_invert(x):
            if isinstance(x, float):
                x = int(x)
            return ~x & 0xff

        # ~image is trimmed to image max so it's hard to test for all formats
        # just test uchar
        self.run_unary(self.all_images, my_invert,
                       fmt=[pyvips.BandFormat.UCHAR])

    # test the rest of VipsArithmetic

    def test_avg(self):
        im = pyvips.Image.black(50, 100)
        test = im.insert(im + 100, 50, 0, expand=True)

        for fmt in all_formats:
            assert pytest.approx(test.cast(fmt).avg()) == 50

    def test_deviate(self):
        im = pyvips.Image.black(50, 100)
        test = im.insert(im + 100, 50, 0, expand=True)

        for fmt in noncomplex_formats:
            assert pytest.approx(test.cast(fmt).deviate(), abs=0.01) == 50

    def test_polar(self):
        im = pyvips.Image.black(100, 100) + 100
        im = im.complexform(im)

        im = im.polar()

        assert pytest.approx(im.real().avg()) == 100 * 2 ** 0.5
        assert pytest.approx(im.imag().avg()) == 45

    def test_rect(self):
        im = pyvips.Image.black(100, 100)
        im = (im + 100 * 2 ** 0.5).complexform(im + 45)

        im = im.rect()

        assert pytest.approx(im.real().avg()) == 100
        assert pytest.approx(im.imag().avg()) == 100

    def test_conjugate(self):
        im = pyvips.Image.black(100, 100) + 100
        im = im.complexform(im)

        im = im.conj()

        assert pytest.approx(im.real().avg()) == 100
        assert pytest.approx(im.imag().avg()) == -100

    def test_histfind(self):
        im = pyvips.Image.black(50, 100)
        test = im.insert(im + 10, 50, 0, expand=True)

        for fmt in all_formats:
            hist = test.cast(fmt).hist_find()
            assert_almost_equal_objects(hist(0, 0), [5000])
            assert_almost_equal_objects(hist(10, 0), [5000])
            assert_almost_equal_objects(hist(5, 0), [0])

        test = test * [1, 2, 3]

        for fmt in all_formats:
            hist = test.cast(fmt).hist_find(band=0)
            assert_almost_equal_objects(hist(0, 0), [5000])
            assert_almost_equal_objects(hist(10, 0), [5000])
            assert_almost_equal_objects(hist(5, 0), [0])

            hist = test.cast(fmt).hist_find(band=1)
            assert_almost_equal_objects(hist(0, 0), [5000])
            assert_almost_equal_objects(hist(20, 0), [5000])
            assert_almost_equal_objects(hist(5, 0), [0])

    def test_histfind_indexed(self):
        im = pyvips.Image.black(50, 100)
        test = im.insert(im + 10, 50, 0, expand=True)
        index = test // 10

        for x in noncomplex_formats:
            for y in [pyvips.BandFormat.UCHAR, pyvips.BandFormat.USHORT]:
                a = test.cast(x)
                b = index.cast(y)
                hist = a.hist_find_indexed(b)

                assert_almost_equal_objects(hist(0, 0), [0])
                assert_almost_equal_objects(hist(1, 0), [50000])

    def test_histfind_ndim(self):
        im = pyvips.Image.black(100, 100) + [1, 2, 3]

        for fmt in noncomplex_formats:
            hist = im.cast(fmt).hist_find_ndim()

            assert_almost_equal_objects(hist(0, 0)[0], 10000)
            assert_almost_equal_objects(hist(5, 5)[5], 0)

            hist = im.cast(fmt).hist_find_ndim(bins=1)

            assert_almost_equal_objects(hist(0, 0)[0], 10000)
            assert hist.width == 1
            assert hist.height == 1
            assert hist.bands == 1

    def test_hough_circle(self):
        test = pyvips.Image.black(100, 100).draw_circle(100, 50, 50, 40)

        for fmt in all_formats:
            im = test.cast(fmt)
            hough = im.hough_circle(min_radius=35, max_radius=45)

            v, x, y = hough.maxpos()
            vec = hough(x, y)
            r = vec.index(v) + 35

            assert pytest.approx(x) == 50
            assert pytest.approx(y) == 50
            assert pytest.approx(r) == 40

    def test_hough_line(self):
        # hough_line changed the way it codes parameter space (again) in 8.17
        test = pyvips.Image.black(100, 100).draw_line(100, 10, 90, 90, 10)

        for fmt in all_formats:
            im = test.cast(fmt)
            hough = im.hough_line()

            v, x, y = hough.maxpos()

            angle = 180.0 * x // hough.width
            distance = test.height * y // hough.height

            assert pytest.approx(angle) == 45
            assert pytest.approx(distance) == 75

    def test_sin(self):
        def my_sin(x):
            if isinstance(x, pyvips.Image):
                return x.sin()
            else:
                return math.sin(math.radians(x))

        self.run_unary(self.all_images, my_sin, fmt=noncomplex_formats)

    def test_cos(self):
        def my_cos(x):
            if isinstance(x, pyvips.Image):
                return x.cos()
            else:
                return math.cos(math.radians(x))

        self.run_unary(self.all_images, my_cos, fmt=noncomplex_formats)

    def test_tan(self):
        def my_tan(x):
            if isinstance(x, pyvips.Image):
                return x.tan()
            else:
                return math.tan(math.radians(x))

        self.run_unary(self.all_images, my_tan, fmt=noncomplex_formats)

    def test_asin(self):
        def my_asin(x):
            if isinstance(x, pyvips.Image):
                return x.asin()
            else:
                return math.degrees(math.asin(x))

        im = (pyvips.Image.black(100, 100) + [1, 2, 3]) / 3.0
        self.run_unary([im], my_asin, fmt=noncomplex_formats)

    def test_acos(self):
        def my_acos(x):
            if isinstance(x, pyvips.Image):
                return x.acos()
            else:
                return math.degrees(math.acos(x))

        im = (pyvips.Image.black(100, 100) + [1, 2, 3]) / 3.0
        self.run_unary([im], my_acos, fmt=noncomplex_formats)

    def test_atan(self):
        def my_atan(x):
            if isinstance(x, pyvips.Image):
                return x.atan()
            else:
                return math.degrees(math.atan(x))

        im = (pyvips.Image.black(100, 100) + [1, 2, 3]) / 3.0
        self.run_unary([im], my_atan, fmt=noncomplex_formats)

    def test_sinh(self):
        def my_sinh(x):
            if isinstance(x, pyvips.Image):
                return x.sinh()
            else:
                return math.sinh(x)

        self.run_unary(self.all_images, my_sinh, fmt=noncomplex_formats)

    def test_cosh(self):
        def my_cosh(x):
            if isinstance(x, pyvips.Image):
                return x.cosh()
            else:
                return math.cosh(x)

        self.run_unary(self.all_images, my_cosh, fmt=noncomplex_formats)

    def test_tanh(self):
        def my_tanh(x):
            if isinstance(x, pyvips.Image):
                return x.tanh()
            else:
                return math.tanh(x)

        self.run_unary(self.all_images, my_tanh, fmt=noncomplex_formats)

    def test_asinh(self):
        def my_asinh(x):
            if isinstance(x, pyvips.Image):
                return x.asinh()
            else:
                return math.asinh(x)

        im = (pyvips.Image.black(100, 100) + [4, 5, 6]) / 3.0
        self.run_unary([im], my_asinh, fmt=noncomplex_formats)

    def test_acosh(self):
        def my_acosh(x):
            if isinstance(x, pyvips.Image):
                return x.acosh()
            else:
                return math.acosh(x)

        im = (pyvips.Image.black(100, 100) + [4, 5, 6]) / 3.0
        self.run_unary([im], my_acosh, fmt=noncomplex_formats)

    def test_atanh(self):
        def my_atanh(x):
            if isinstance(x, pyvips.Image):
                return x.atanh()
            else:
                return math.atanh(x)

        im = (pyvips.Image.black(100, 100) + [0, 1, 2]) / 3.0
        self.run_unary([im], my_atanh, fmt=noncomplex_formats)

    def test_atan2(self):
        def my_atan2(x, y):
            if isinstance(x, pyvips.Image):
                return x.atan2(y)
            else:
                return math.degrees(math.atan2(x[0], y[0]))

        im = (pyvips.Image.black(100, 100) + [1, 2, 3]) / 3.0
        self.run_binary(im, my_atan2, fmt=noncomplex_formats)

    def test_log(self):
        def my_log(x):
            if isinstance(x, pyvips.Image):
                return x.log()
            else:
                return math.log(x)

        self.run_unary(self.all_images, my_log, fmt=noncomplex_formats)

    def test_log10(self):
        def my_log10(x):
            if isinstance(x, pyvips.Image):
                return x.log10()
            else:
                return math.log10(x)

        self.run_unary(self.all_images, my_log10, fmt=noncomplex_formats)

    def test_exp(self):
        def my_exp(x):
            if isinstance(x, pyvips.Image):
                return x.exp()
            else:
                return math.exp(x)

        self.run_unary(self.all_images, my_exp, fmt=noncomplex_formats)

    def test_exp10(self):
        def my_exp10(x):
            if isinstance(x, pyvips.Image):
                return x.exp10()
            else:
                return math.pow(10, x)

        self.run_unary(self.all_images, my_exp10, fmt=noncomplex_formats)

    def test_floor(self):
        def my_floor(x):
            if isinstance(x, pyvips.Image):
                return x.floor()
            else:
                return math.floor(x)

        self.run_unary(self.all_images, my_floor)

    def test_ceil(self):
        def my_ceil(x):
            if isinstance(x, pyvips.Image):
                return x.ceil()
            else:
                return math.ceil(x)

        self.run_unary(self.all_images, my_ceil)

    def test_rint(self):
        def my_rint(x):
            if isinstance(x, pyvips.Image):
                return x.rint()
            else:
                return round(x)

        self.run_unary(self.all_images, my_rint)

    def test_sign(self):
        def my_sign(x):
            if isinstance(x, pyvips.Image):
                return x.sign()
            else:
                if x > 0:
                    return 1
                elif x < 0:
                    return -1
                else:
                    return 0

        self.run_unary(self.all_images, my_sign)

    def test_max(self):
        test = pyvips.Image.black(100, 100).draw_rect(100, 40, 50, 1, 1)

        for fmt in all_formats:
            v = test.cast(fmt).max()

            assert pytest.approx(v) == 100
            v, x, y = test.cast(fmt).maxpos()
            assert pytest.approx(v) == 100
            assert pytest.approx(x) == 40
            assert pytest.approx(y) == 50

    def test_min(self):
        test = (pyvips.Image.black(100, 100) + 100).draw_rect(0, 40, 50, 1, 1)

        for fmt in all_formats:
            v = test.cast(fmt).min()

            assert pytest.approx(v) == 0
            v, x, y = test.cast(fmt).minpos()
            assert pytest.approx(v) == 0
            assert pytest.approx(x) == 40
            assert pytest.approx(y) == 50

    def test_measure(self):
        im = pyvips.Image.black(50, 50)
        test = im.insert(im + 10, 50, 0, expand=True)

        for x in noncomplex_formats:
            a = test.cast(x)
            matrix = a.measure(2, 1)
            [p1] = matrix(0, 0)
            [p2] = matrix(0, 1)

            assert pytest.approx(p1) == 0
            assert pytest.approx(p2) == 10

    def test_find_trim(self):
        im = pyvips.Image.black(50, 60) + 100
        test = im.embed(10, 20, 200, 300, extend="white")

        for x in unsigned_formats + float_formats:
            a = test.cast(x)
            left, top, width, height = a.find_trim()

            assert left == 10
            assert top == 20
            assert width == 50
            assert height == 60

        test_rgb = test.bandjoin([test, test])
        left, top, width, height = test_rgb.find_trim(line_art=True,
                                                      background=[255, 255,
                                                                  255])
        assert left == 10
        assert top == 20
        assert width == 50
        assert height == 60

    def test_profile(self):
        test = pyvips.Image.black(100, 100).draw_rect(100, 40, 50, 1, 1)

        for fmt in noncomplex_formats:
            columns, rows = test.cast(fmt).profile()

            v, x, y = columns.minpos()
            assert pytest.approx(v) == 50
            assert pytest.approx(x) == 40
            assert pytest.approx(y) == 0

            v, x, y = rows.minpos()
            assert pytest.approx(v) == 40
            assert pytest.approx(x) == 0
            assert pytest.approx(y) == 50

    def test_project(self):
        im = pyvips.Image.black(50, 50)
        test = im.insert(im + 10, 50, 0, expand=True)

        for fmt in noncomplex_formats:
            columns, rows = test.cast(fmt).project()

            assert_almost_equal_objects(columns(10, 0), [0])
            assert_almost_equal_objects(columns(70, 0), [50 * 10])

            assert_almost_equal_objects(rows(0, 10), [50 * 10])

    def test_stats(self):
        im = pyvips.Image.black(50, 50)
        test = im.insert(im + 10, 50, 0, expand=True)

        for x in noncomplex_formats:
            a = test.cast(x)
            matrix = a.stats()

            assert_almost_equal_objects(matrix(0, 0), [a.min()])
            assert_almost_equal_objects(matrix(1, 0), [a.max()])
            assert_almost_equal_objects(matrix(2, 0), [50 * 50 * 10])
            assert_almost_equal_objects(matrix(3, 0), [50 * 50 * 100])
            assert_almost_equal_objects(matrix(4, 0), [a.avg()])
            assert_almost_equal_objects(matrix(5, 0), [a.deviate()])

            assert_almost_equal_objects(matrix(0, 1), [a.min()])
            assert_almost_equal_objects(matrix(1, 1), [a.max()])
            assert_almost_equal_objects(matrix(2, 1), [50 * 50 * 10])
            assert_almost_equal_objects(matrix(3, 1), [50 * 50 * 100])
            assert_almost_equal_objects(matrix(4, 1), [a.avg()])
            assert_almost_equal_objects(matrix(5, 1), [a.deviate()])

    def test_sum(self):
        for fmt in all_formats:
            im = pyvips.Image.black(50, 50)
            im2 = [(im + x).cast(fmt) for x in range(0, 100, 10)]
            im3 = pyvips.Image.sum(im2)
            assert pytest.approx(im3.max()) == sum(range(0, 100, 10))

    def test_clamp(self):
        for fmt in noncomplex_formats:
            im = self.colour
            for x in range(0, 100, 10):
                im2 = (im + x).cast(fmt)
                im3 = im2.clamp()
                assert im3.max() <= 1.0
                assert im3.min() >= 0.0

                im3 = im2.clamp(min=14, max=45)
                assert im3.max() <= 45
                assert im3.min() >= 14

    def test_minpair(self):
        for fmt in noncomplex_formats:
            im = self.colour
            for x in range(0, 100, 10):
                im2 = ((im - x) * 5).cast(fmt)
                im3 = im2.minpair(im)
                im4 = (im2 < im).ifthenelse(im2, im)
                assert (im3 - im4).abs().max() == 0

    def test_maxpair(self):
        for fmt in noncomplex_formats:
            im = self.colour
            for x in range(0, 100, 10):
                im2 = ((im - x) * 5).cast(fmt)
                im3 = im2.maxpair(im)
                im4 = (im2 > im).ifthenelse(im2, im)
                assert (im3 - im4).abs().max() == 0


if __name__ == '__main__':
    pytest.main()



# vim: set fileencoding=utf-8 :
import operator
import pytest
from functools import reduce

import pyvips
from helpers import *


# point convolution
def conv(image, mask, x_position, y_position):
    s = 0.0
    for x in range(0, mask.width):
        for y in range(0, mask.height):
            m = mask(x, y)
            i = image(x + x_position, y + y_position)
            p = run_fn2(operator.mul, m, i)
            s = run_fn2(operator.add, s, p)

    return run_fn2(operator.truediv, s, mask.scale)


def compass(image, mask, x_position, y_position, n_rot, fn):
    acc = []
    for i in range(0, n_rot):
        result = conv(image, mask, x_position, y_position)
        result = run_fn(abs, result)
        acc.append(result)
        mask = mask.rot45()

    return reduce(lambda a, b: run_fn2(fn, a, b), acc)


class TestConvolution:
    @classmethod
    def setup_class(cls):
        im = pyvips.Image.mask_ideal(100, 100, 0.5, reject=True, optical=True)
        cls.colour = im * [1, 2, 3] + [2, 3, 4]
        cls.colour = cls.colour.copy(interpretation=pyvips.Interpretation.SRGB)
        cls.mono = cls.colour.extract_band(1)
        cls.mono = cls.mono.copy(interpretation=pyvips.Interpretation.B_W)
        cls.all_images = [cls.mono, cls.colour]
        cls.sharp = pyvips.Image.new_from_array([[-1, -1, -1],
                                                 [-1, 16, -1],
                                                 [-1, -1, -1]], scale=8)
        cls.blur = pyvips.Image.new_from_array([[1, 1, 1],
                                                [1, 1, 1],
                                                [1, 1, 1]], scale=9)
        cls.line = pyvips.Image.new_from_array([[1, 1, 1],
                                                [-2, -2, -2],
                                                [1, 1, 1]])
        cls.sobel = pyvips.Image.new_from_array([[1, 2, 1],
                                                 [0, 0, 0],
                                                 [-1, -2, -1]])
        cls.all_masks = [cls.sharp, cls.blur, cls.line, cls.sobel]

    @classmethod
    def teardown_class(cls):
        cls.colour = None
        cls.mono = None
        cls.all_images = None
        cls.sharp = None
        cls.blur = None
        cls.line = None
        cls.sobel = None
        cls.all_masks = None

    def test_conv(self):
        for im in self.all_images:
            for msk in self.all_masks:
                for prec in [pyvips.Precision.INTEGER, pyvips.Precision.FLOAT]:
                    convolved = im.conv(msk, precision=prec)

                    result = convolved(25, 50)
                    true = conv(im, msk, 24, 49)
                    assert_almost_equal_objects(result, true)

                    result = convolved(50, 50)
                    true = conv(im, msk, 49, 49)
                    assert_almost_equal_objects(result, true)

        # https://github.com/libvips/libvips/issues/4955
        # (cast to ushort to avoid vector path)
        im = pyvips.Image.mask_ideal(100, 100, 0.5, reject=True, optical=True, uchar=True).cast('ushort')
        convolved = im.conv(self.sharp, precision=pyvips.Precision.INTEGER)
        assert convolved(24, 49) == [0]

    # don't test conva, it's still not done
    def dont_est_conva(self):
        for im in self.all_images:
            for msk in self.all_masks:
                print("msk:")
                msk.matrixprint()
                print("im.bands = %s" % im.bands)

                convolved = im.conv(msk,
                                    precision=pyvips.Precision.APPROXIMATE)

                result = convolved(25, 50)
                true = conv(im, msk, 24, 49)
                print("result = %s, true = %s" % (result, true))
                assert_less_threshold(result, true, 5)

                result = convolved(50, 50)
                true = conv(im, msk, 49, 49)
                print("result = %s, true = %s" % (result, true))
                assert_less_threshold(result, true, 5)

    def test_compass(self):
        for im in self.all_images:
            for msk in self.all_masks:
                for prec in [pyvips.Precision.INTEGER, pyvips.Precision.FLOAT]:
                    for times in range(1, 4):
                        convolved = im.compass(msk,
                                               times=times,
                                               angle=pyvips.Angle45.D45,
                                               combine=pyvips.Combine.MAX,
                                               precision=prec)

                        result = convolved(25, 50)
                        true = compass(im, msk, 24, 49, times, max)
                        assert_almost_equal_objects(result, true)

        for im in self.all_images:
            for msk in self.all_masks:
                for prec in [pyvips.Precision.INTEGER, pyvips.Precision.FLOAT]:
                    for times in range(1, 4):
                        convolved = im.compass(msk,
                                               times=times,
                                               angle=pyvips.Angle45.D45,
                                               combine=pyvips.Combine.SUM,
                                               precision=prec)

                        result = convolved(25, 50)
                        true = compass(im, msk, 24, 49, times, operator.add)
                        assert_almost_equal_objects(result, true)

    def test_convsep(self):
        for im in self.all_images:
            for prec in [pyvips.Precision.INTEGER, pyvips.Precision.FLOAT]:
                gmask = pyvips.Image.gaussmat(2, 0.1,
                                              precision=prec)
                gmask_sep = pyvips.Image.gaussmat(2, 0.1,
                                                  separable=True,
                                                  precision=prec)

                assert gmask.width == gmask.height
                assert gmask_sep.width == gmask.width
                assert gmask_sep.height == 1

                a = im.conv(gmask, precision=prec)
                b = im.convsep(gmask_sep, precision=prec)

                a_point = a(25, 50)
                b_point = b(25, 50)

                assert_almost_equal_objects(a_point, b_point, threshold=0.1)

    def test_fastcor(self):
        for im in self.all_images:
            for fmt in noncomplex_formats:
                small = im.crop(20, 45, 10, 10).cast(fmt)
                cor = im.fastcor(small)
                v, x, y = cor.minpos()

                assert v == 0
                assert x == 25
                assert y == 50

    def test_spcor(self):
        for im in self.all_images:
            for fmt in noncomplex_formats:
                small = im.crop(20, 45, 10, 10).cast(fmt)
                cor = im.spcor(small)
                v, x, y = cor.maxpos()

                assert v == 1.0
                assert x == 25
                assert y == 50

    def test_gaussblur(self):
        for im in self.all_images:
            for prec in [pyvips.Precision.INTEGER, pyvips.Precision.FLOAT]:
                for i in range(5, 10):
                    sigma = i / 5.0
                    gmask = pyvips.Image.gaussmat(sigma, 0.2,
                                                  precision=prec)

                    a = im.conv(gmask, precision=prec)
                    b = im.gaussblur(sigma, min_ampl=0.2, precision=prec)

                    a_point = a(25, 50)
                    b_point = b(25, 50)

                    assert_almost_equal_objects(a_point, b_point,
                                                threshold=0.1)

    def test_sharpen(self):
        for im in self.all_images:
            for fmt in noncomplex_formats:
                # old vipses used "radius", check that that still works
                sharp = im.sharpen(radius=5)

                for sigma in [0.5, 1, 1.5, 2]:
                    im = im.cast(fmt)
                    sharp = im.sharpen(sigma=sigma)

                    # hard to test much more than this
                    assert im.width == sharp.width
                    assert im.height == sharp.height

                    # if m1 and m2 are zero, sharpen should do nothing
                    sharp = im.sharpen(sigma=sigma, m1=0, m2=0)
                    sharp = sharp.colourspace(im.interpretation)
                    # print("testing sig = %g" % sigma)
                    # print("testing fmt = %s" % fmt)
                    # print("max diff = %g" % (im - sharp).abs().max())
                    assert (im - sharp).abs().max() == 0


if __name__ == '__main__':
    pytest.main()



# vim: set fileencoding=utf-8 :
import pytest

import pyvips
from helpers import *


class TestMosaicing:

    def test_lrmerge(self):
        left = pyvips.Image.new_from_file(MOSAIC_FILES[0])
        right = pyvips.Image.new_from_file(MOSAIC_FILES[1])
        join = left.merge(right, 'horizontal', 10 - left.width, 0)

        assert join.width == left.width + right.width - 10
        assert join.height == max(left.height, right.height)
        assert join.bands == 1

    def test_tbmerge(self):
        top = pyvips.Image.new_from_file(MOSAIC_FILES[0])
        bottom = pyvips.Image.new_from_file(MOSAIC_FILES[2])
        join = top.merge(bottom, 'vertical', 0, 10 - top.height)

        assert join.width == max(top.width, bottom.width)
        assert join.height == top.height + bottom.height - 10
        assert join.bands == 1

    def test_lrmosaic(self):
        left = pyvips.Image.new_from_file(MOSAIC_FILES[0])
        right = pyvips.Image.new_from_file(MOSAIC_FILES[1])
        join = left.mosaic(right, 'horizontal', left.width - 30, 0, 30, 0)

        assert join.width == 1014
        assert join.height == 379
        assert join.bands == 1

    def test_tbmosaic(self):
        top = pyvips.Image.new_from_file(MOSAIC_FILES[0])
        bottom = pyvips.Image.new_from_file(MOSAIC_FILES[2])
        join = top.mosaic(bottom, 'vertical', 0, top.height - 30, 0, 30)

        assert join.width == 542
        assert join.height == 688
        assert join.bands == 1

    def test_mosaic(self):
        # ported from https://github.com/libvips/nip2/tree/master/share/nip2/data/examples/1_point_mosaic

        mosaiced_image = None

        for i in range(0, len(MOSAIC_FILES), 2):
            files = MOSAIC_FILES[i:i + 2]
            marks = MOSAIC_MARKS[i:i + 2]

            im = pyvips.Image.new_from_file(files[0])
            sec_im = pyvips.Image.new_from_file(files[1])
            horizontal_part = im.mosaic(sec_im,
                    pyvips.Direction.HORIZONTAL,
                    marks[0][0], marks[0][1],
                    marks[1][0], marks[1][1])

            if mosaiced_image is None:
                mosaiced_image = horizontal_part
            else:
                vertical_marks = MOSAIC_VERTICAL_MARKS[i - 2:i]
                mosaiced_image = mosaiced_image.mosaic(horizontal_part,
                       pyvips.Direction.VERTICAL,
                       vertical_marks[1][0], vertical_marks[1][1],
                       vertical_marks[0][0], vertical_marks[0][1])

        # Uncomment to see output file
        #mosaiced_image.write_to_file('after.jpg')

        # hard to test much more than this
        assert mosaiced_image.width == 1005
        assert mosaiced_image.height == 1295
        assert mosaiced_image.interpretation == pyvips.Interpretation.B_W
        assert mosaiced_image.bands == 1

    def test_globalbalance(self):
        mosaiced_image = None

        for i in range(0, len(MOSAIC_FILES), 2):
            files = MOSAIC_FILES[i:i + 2]
            marks = MOSAIC_MARKS[i:i + 2]

            im = pyvips.Image.new_from_file(files[0])
            sec_im = pyvips.Image.new_from_file(files[1])
            horizontal_part = im.mosaic(sec_im,
                    pyvips.Direction.HORIZONTAL,
                    marks[0][0], marks[0][1],
                    marks[1][0], marks[1][1])

            if mosaiced_image is None:
                mosaiced_image = horizontal_part
            else:
                vertical_marks = MOSAIC_VERTICAL_MARKS[i - 2:i]
                mosaiced_image = mosaiced_image.mosaic(horizontal_part,
                       pyvips.Direction.VERTICAL,
                       vertical_marks[1][0], vertical_marks[1][1],
                       vertical_marks[0][0], vertical_marks[0][1])

        mosaiced_image = mosaiced_image.globalbalance()

        # Uncomment to see output file
        #mosaiced_image.write_to_file('after.jpg')

        # hard to test much more than this
        assert mosaiced_image.width == 1005
        assert mosaiced_image.height == 1295
        assert mosaiced_image.interpretation == pyvips.Interpretation.B_W
        assert mosaiced_image.format == pyvips.BandFormat.FLOAT
        assert mosaiced_image.bands == 1

if __name__ == '__main__':
    pytest.main()
